/* compile:
icc -g -O -Qoption,c,-ip_ninl_max_stats=500 -tpp6 -xK timer.o repro-6.c -lpthread -DVERBOSE
*/

/************************************************************************************
 
    REPRO -- A PROGRAM FOR FINDING DISTANTLY REPEATED FRAGMENTS IN A PROTEIN SEQUENCE 

    Copyright (c) 1992  Jaap Heringa - EMBL Heidelberg
    Copyright (c) 1999  Jaap Heringa - NIMR London

    Jaap Heringa (jhering@nimr.mrc.ac.uk)
    National Institute for Medical Research (NIMR)
    The Ridgeway
    Mill Hill
    London  NW7 1AA
    United Kingdom

    -- Run program without parameters on the command line to find out about the REPRO parameters 
    -- A Web server is available at http://mathbio.nimr.mrc.ac.uk

    Please cite: 
      Heringa J & Argos P (1993) A method to recognize distant repeats in protein sequences. 
      Proteins Struct. Func. Genet. 17, 391-411.

    And when using the web server: 
      George RA & Heringa J (2000) Repro: a web server for delineating protein sequence repeats. 
      Trends Biochem. Sci. 25, 515-517.


Example of file format for residue exchange matrix needed by program (pam250). 
Create pam250.mat file and paste in the section in between the two dashed lines: 
-----------------------
ARNDCQEGHILKMFPSTWYVBZX
  2
 -2  6
  0  0  2
  0 -1  2  4
 -2 -4 -4 -5 12
  0  1  1  2 -5  4
  0 -1  1  3 -5  2  4
  1 -3  0  1 -3 -1  0  5
 -1  2  2  1 -3  3  1 -2  6
 -1 -2 -2 -2 -2 -2 -2 -3 -2  5
 -2 -3 -3 -4 -6 -2 -3 -4 -2  2  6
 -1  3  1  0 -5  1  0 -2  0 -2 -3  5
 -1  0 -2 -3 -5 -1 -2 -3 -2  2  4  0  6
 -4 -4 -4 -6 -4 -5 -5 -5 -2  1  2 -5  0  9
  1  0 -1 -1 -3  0 -1 -1  0 -2 -3 -1 -2 -5  6
  1  0  1  0  0 -1  0  1 -1 -1 -3  0 -2 -3  1  2
  1 -1  0  0 -2 -1  0  0 -1  0 -2  0 -1 -3  0  1  3
 -6  2 -4 -7 -8 -5 -7 -7 -3 -5 -2 -3 -4  0 -6 -2 -5 17
 -3 -4 -2 -4  0 -4 -4 -5  0 -1 -1 -4 -2  7 -5 -3 -3  0 10
  0 -2 -2 -2 -2 -2 -2 -1 -2  4  2 -2  2 -1 -1 -1  0 -6 -2  4
  0 -1  2  3 -4  1  2  0  1 -2 -3  1 -2 -5 -1  0  0 -5 -3 -2  2
  0  0  1  3 -5  3  3 -1  2 -2 -3  0 -2 -5  0  0 -1 -6 -4 -2  2  3
  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0
-----------------------
************************************************************************************/

#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined TIMER
#include "timer.h"
#endif

/*
#define test
#define test2 
#define SHOWCLUSTER 
*/

#if !defined DATA_TYPE
#define DATA_TYPE SHORT_TYPE
#endif

#if defined __INTEL_COMPILER
#if defined P3
#if !defined NO_SSE && !defined SSE && !defined SSE2
#define SSE
#endif
#if !defined CACHE_SIZE
#define CACHE_SIZE 16384
#endif
#endif

#if defined P4
#if !defined NO_SSE && !defined SSE && !defined SSE2
#define SSE2
#endif
#if !defined CACHE_SIZE
#define CACHE_SIZE 8192
#endif
#endif
#endif

#if !defined CACHE_SIZE
#define CACHE_SIZE 16384
#endif

#if !defined LINE_SIZE
#define LINE_SIZE  (CACHE_SIZE / sizeof(block_type) / 3)
#endif

#define SHORT_TYPE 0
#define FLOAT_TYPE 1

#if DATA_TYPE == SHORT_TYPE
typedef short score_type;
#define MAX_SCORE 32767
#define MPI_DATA_TYPE	MPI_SHORT
#elif DATA_TYPE == FLOAT_TYPE
typedef float score_type;
#define MAX_SCORE INFINITY
#define MPI_DATA_TYPE	MPI_FLOAT
#endif

#if defined SSE || defined SSE2
#include <mmintrin.h>
#include <xmmintrin.h>
#endif

#if defined SSE2
#include <emmintrin.h>
#endif

#if defined MULTI_THREADING
#define __USE_GNU
#include <pthread.h>
#endif

#if defined MULTI_PROCESSOR
#include <mpi.h>
#endif

#if 1
#  if DATA_TYPE == SHORT_TYPE
#    if defined SSE2
       typedef __m128i    block_type;
#      define BLOCK_SIZE 8
#    elif defined SSE
       typedef __m64      block_type;
#      define BLOCK_SIZE 4
#    else
       typedef score_type block_type;
#      define BLOCK_SIZE 1
#    endif
#  elif DATA_TYPE == FLOAT_TYPE
#    if defined SSE || defined SSE2
       typedef __m128     block_type;
#      define BLOCK_SIZE 4
#    else
       typedef score_type block_type;
#      define BLOCK_SIZE 1
#    endif
#  endif
#endif

#if DATA_TYPE == FLOAT_TYPE
typedef float	    score_vector_type;
#else
typedef signed char score_vector_type;
#endif


#if !defined _cplusplus
typedef char bool;
#define false 0
#define true  1
#endif

#if !defined __GNUC__ && !defined __INTEL_COMPILER
#define __inline__ 
#endif

#if defined MULTI_THREADING
int nr_threads = 1;
#endif

#if defined MULTI_PROCESSOR
int node_id, master_node_id, nr_nodes;
#endif


#define LF 10

#define VMULT  1
#define RVMULT (1.0 / VMULT)
#define GAPYOFFS  -50000
#define ALLSTARTLIM 1000
#define TEMPSTARTLIM 2000
#define UNIX
#define PWLINLEN 100
#define SALINLEN 10000


#define DIAGONAL 1
#define LEFT	 2
#define UP	 3

#ifdef UNIX
#define SIMFILE    "/home/www/htdocs/Services/argos/repro/gribskov.mat"
#endif

#ifdef VAX
#define SIMFILE    "$u:[heringa.poly]grb.sim2"
#endif

/* macro's */
#define _ScrGapx(k, script, length) { *(++script) = - (k); ++length; }
#define _ScrGapy(k, script, length) { *(++script) = GAPYOFFS - (k); ++length; }
#define _ScrDiag(script, length) \
{ if (*script <= 0) \
    {*(++script) = 1; ++length;} \
  else \
    (*script)++;}

#define _Abs(v) ((v) >= 0)? (v) : -(v)

#define _Min(a, b)        (((a) > (b)) ? (b) : (a))
#define _Min3(a, b, c)    (((a) > (_Min(b, c))) ? (_Min(b, c)) : (a))
#define _Max(a, b)        (((a) > (b)) ? (a) : (b))
#define _Max3(a, b, c)    (((a) > (_Max((b), (c)))) ? (a) : (_Max((b), (c))))
#define _Max4(a, b, c, d) (((a) > (_Max3((b), (c), (d)))) ? (a) : (_Max3((b), (c), (d))))
#define alloc_1d(ptr, type, n, msg) { \
    if (n == 0) { \
    fprintf(stdout,"  Error: zero allocation requested for: %s\n", msg); \
    exit(0); \
    }; \
    if ((ptr = (type *) ckcalloc((n), sizeof(type), msg)) == NULL) { \
    fprintf(stdout,"  Error: no storage for %s", msg); \
    exit(0); \
    } \
}
#define alloc_2d(ptr, type, n1, n2, msg) { \
    int loop; \
    if (n1 == 0 || n2 == 0) { \
    fprintf(stdout,"  Error: zero allocation [%d, %d] requested for: %s\n", msg, n1, n2); \
    exit(0); \
    } \
    if ((ptr = (type **) ckcalloc((n1), sizeof(type *), msg)) == NULL) { \
	fprintf(stdout,"  Error: no storage for ptr block: %s", msg); \
	exit(0); \
	} \
    if ((ptr[0] = (type *) ckcalloc((n1)*(n2), sizeof(type), msg)) == NULL) { \
	fprintf(stdout, "  Error: no storage for data block: %s", msg); \
	exit(0); \
	} \
    for (loop = 1; loop < n1; loop++) \
      ptr[loop] = ptr[loop-1] + (n2); \
}
#define alloc_3d(ptr, type, n1, n2, n3, msg) { \
    int loop_1, loop_2; \
    if (n1 == 0 || n2 == 0 || n3 == 0) { \
    fprintf(stdout,"  Error: zero allocation [%d, %d, %d] requested for: %s\n", msg, n1, n2, n3); \
    exit(0); \
    } \
    if ((ptr = (type ***) ckcalloc((n1), sizeof(type **), msg)) == NULL) { \
       fprintf(stdout,"  Error: no storage for ptr block: %s", msg); \
       exit(0); \
       } \
    if ((ptr[0] = (type **) ckcalloc((n1)*(n2), sizeof(type *), msg)) == NULL) { \
       fprintf(stdout, "  Error: no storage for data block: %s", msg); \
       exit(0); \
       } \
    if ((ptr[0][0] = (type *) ckcalloc((n1)*(n2)*(n3), sizeof(type), msg)) == NULL) {\
       fprintf(stdout, "  Error: no storage for data block: %s", msg); \
       exit(0); \
       } \
    for (loop_1 = 0; loop_1 < n1; loop_1++) { \
       if (loop_1 > 0) { \
          ptr[loop_1] = ptr[loop_1-1] + (n2); \
          ptr[loop_1][0] = ptr[loop_1-1][loop_2-1] + (n3); \
          } \
       for (loop_2 = 1; loop_2 < n2; loop_2++) { \
           ptr[loop_1][loop_2] = ptr[loop_1][loop_2-1] + (n3); \
           } \
       } \
}

#define free_1d(ptr) { ckfree(ptr);}
#define free_2d(ptr) { ckfree(ptr[0]); ckfree(ptr);}
#define free_3d(ptr) { ckfree(ptr[0][0]); ckfree(ptr[0]); ckfree(ptr);}



/* typedef's */
typedef struct script
{
    long yoffs, x1, y1, x2, y2, len, score, *lis, size, coordnum;
    struct script *next;
}
S_Script;

typedef struct graph
{
    long graphid, subgraphid, subgraphnum;
}
S_Subgraph;

struct long3
{
    long i, j, val;
};


/* function prototypes */
void Def_pam (long V[128][128], char alpha[]);
void *ckalloc (size_t amount, const char *msg);	/* space-allocating function  */
void *ckcalloc (size_t amount, size_t len, const char *msg);	/* calloc + test    */
void ckfree(void *);
void fatal(const char *msg);
FILE *ckopen (const char *name, const char *mode);	/* file opening function      */
void AliAddMatrix (char A[], long M, long V[128][128]);
float Ali2 (long **m, long x_n, long y_n, long gapi, long gapx);
const char *prefix (const char *fn, char *s);
void ScriptLength (S_Script * ptr);
void PrintAli (char A[], S_Script * p, long num, FILE * fp);
void StackAli (char A[], long M, long **ali, S_Script * p, long num,
	       FILE * fp);
void PrintStackAli (long **ali, char A[], long M, long dim, FILE * fp);
long Traceback (long **m, long yoffs, long x, long y, long score,
		long gapi, long gapx, long **coord, char *coordtab,
		long topnum, long *freecoordnum, long M);
void MakeGraphs (char A[], long M, long **ali, S_Script * p,
		 long dim, char *filter, FILE * ofp,
		 long consconn, long minfraglen,
		 char *detfraglen, long virtualvar, char *virtualnodes,
		 char *clanfilter);
void RecStart (long **pali, long index1, long index2, long i, long ii,
	       char *filter, long dim, long M, long **ali, long *wght,
	       long *check);
/*
void RecStart(long **pali, long index1, long index2, long i, char *filter, 
	      long dim, long M, long **ali, long *wght);
*/
void StackAll (long *workcount, long *workwght, long i,
	       char *filter, long dim, long M, long **ali, long *wght);
void RecRealStart (long **pali, long index1, long index2, long i, long ii,
		   long alinum, long **start, char *filter,
		   long dim, long M, long **ali, long *wght, long *check,
		   long virtualvar);
void CopyCloseStarts (long **startgraph, long graphnum, long ***allstart,
		      long *allstartnum, long **addstart,
		      long *addstartnum_ptr, char *filter, long scriptsdim,
		      long testfac, long **scr, long virtualvar,
		      char *virtualnodes);
void CopyTempStart (long **addstart, long *addstartnum_ptr, long ***allstart,
		    long *allstartnum, long scriptnum, long p1, long p2);
void CopyStartPair (long **addstart, long addstartnum, long ***allstart,
		    long *allstartnum, long graphnum);
void PrintGraph (FILE * ofp, long graphsize, long **startgraph,
		 long **subgraph, S_Subgraph * lsubgraph, long ***adjmat,
		 long *aliwght, long *alilenwght, long alinum,
		 long **alistart, long **frags, char *graphtitle,
		 char *detfraglen);
void testdecomp (long node1, long dim, long **adjmat, long *subgraph, long *helpgraph);
void ListSubgraphs (long **subgraph, S_Subgraph * lsubgraph, long graphsize, long subgraphsize, long **startgraph);
void Decompose (long **startgraph, long graphnum, long ***adjmat, long **subgraph);
void PrintFrags (FILE * ofp, char A[], long **frags, S_Subgraph * lsubgraph, long subgraphnum);
void DetFragEnds (long **frags, char *detfraglen, long graphloop, long graphnum, long *locgraph, long *locsubgraph, long subgraphid, long alinum, long **alistart, long *helpali, long **locadjmat);
void MakeAdjMats (long ***allstart, long *allstartnum, long **startgraph, long graphnum, long ***adjmat, long dim, char *filter);
void LengthFilter (long **addstart, long i, long **scr, long minfraglen);
void FileCedricScripts (char A[], long M, S_Script * ptr, char *seq_fn, char *script_ofile);
void PrintIntScripts (long mismatch, S_Script * ptr, char *seq_fn, char *script_ofile, char *comp_mat, long gapi, long gapx);
void PrintScripts (char A[], long mismatch, S_Script * ptr, char *seq_fn, char *script_ofile, char *comp_mat, long gapi, long gapx);
void ProcScript (char A[], S_Script * ptr, FILE * fp, char *seq_fn, long mismatch, char *comp_mat, long gapi, long gapx, long consconn, long minfraglen, char *detfraglen, long virtualvar, char *virtualnodes, char *clanfilter);
void ScriptLength (S_Script * ptr);
void Delete_script (S_Script * ptr, S_Script * pre, S_Script * head, long *freecoordnum, long **coord, char *coordtab, long M);
void read_script_file (char *scrfile, char *subst_fn, long topnum);
void Asks (char *s, char *str);
void Verbose (long argc, char *argv[], long *consconn, long *minfraglen, char *detfraglen, long *virtualvar, char *virtualnodes, char *clanfilter);
void Showparam (const char *name);
void Install_script (S_Script * ptr, long *lscr, long score, long yoffs, long x1, long y1, long x2, long y2, long **coord, char *coordtab, long topnum, long *freecoordnum, long M);
void FillCoord (S_Script * p, long *freecoordnum, long **coord, char *coordtab, long M);
int Test_script (S_Script * ptr, long score, long x, long y, long **coord, char *coordtab, long *freecoordnum, long M);


/* global variables */

S_Script script_h;		/* Script header struct */
static long V[128][128];	/* Converted integer weights  */
char *A;			/* Storing the sequence       */
static long topnum;		/* number of ali script items */
long Q, R;			/* Converted integer weights  */
long mism_val;			/* if 1 then def pam250, 0: matrix, 
				   neg value: mismatch val given */
long M;				/* Sequence lengths and k     */

__inline__ static int min(int a, int b)
{
    return a < b ? a : b;
}


__inline__ static int max(int a, int b)
{
    return a > b ? a : b;
}


#if defined MULTI_PROCESSOR

void comm_init(int *argc, char **argv)
{
    MPI_Init(argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &nr_nodes);
    MPI_Comm_rank(MPI_COMM_WORLD, &node_id);
    MPI_Errhandler_set(MPI_COMM_WORLD, MPI_ERRORS_ARE_FATAL);
    master_node_id = nr_nodes - 1; /* This is specific for DAS-2, where the
				      last hosts have most memory */
}


void comm_end(void)
{
    MPI_Finalize();
}

#endif



void set_identity_matrix(const char *str, long V[128][128])
{
    int   i, j;
    float ms;

    sscanf(str, "%f", &ms);

    if (ms >= 0.0)
	fatal("The mismatch weight should be a negative number");

    mism_val = VMULT * ms;

    for (i = 0; i < 128; i++)
	for (j = 0; j < 128; j++)
	    V[i][j] = i == j ? VMULT : mism_val;
}


void read_sequence(const char *file_name)
{
#if defined MULTI_PROCESSOR
    if (node_id == master_node_id) {
#endif
	int  ch;
	FILE *file = ckopen(file_name, "r");

	for (M = 0; (ch = getc(file)) != EOF;)
	    if (!isspace(ch))
		++ M;

	if (M < 2)
	    fatal("sequence should have at least 2 characters");
	    
	A = ckalloc(M + 1, "sequence");
	rewind(file);

	for (M = 0; (ch = getc(file)) != EOF;)
	    if (!isspace(ch))
		A[M++] = ch;

	A[M] = '\0';
	fprintf(stdout, "sequence from file \"%s\" has length %ld\n", file_name, M);

#if defined MULTI_PROCESSOR
    }

    MPI_Bcast(&M, 1, MPI_INT, master_node_id, MPI_COMM_WORLD);

    if (node_id != master_node_id)
	A = ckalloc(M + 1, "sequence");

    MPI_Bcast(A, M + 1, MPI_CHAR, master_node_id, MPI_COMM_WORLD);
#endif
}


void do_work(int argc, char *argv[])
{
    /* *** for www server ***
       char err_file_name[200];
       char out_file_name[200];
     */
    char argv_1[200];		/* to change argv[1]: argv_1 kept (lazy) */

    float ms;			/* User-supplied weights      */
    int  i, j;
    FILE *Sp;
    char alph[129];		/* alphabet */
    char *s, act;		/* alphabet */
    long size;			/* size of alphabet */
    char iscr_fn[233];		/* internal scripts output file name */
    char scr_fn[233];		/* scripts output file name */
    float q, r;			/* gap penalties */
    char subst_fn[200];		/* name of Dayhoff file */
    long paroffset;		/* offset in case of (a) ali only */
    long consconn = 3;		/* number of edges for consistent
				   connection in graph */
    long minfraglen = 1;	/* minimum length for fragments
				   starting just before end of
				   top-alignment */
    char detfraglen[10];	/* determination of fragment ends */
    long virtualvar = 2;	/* length variance for virtual
				   startsites */
    char virtualnodes[81];	/* indicator of start-sites to be
				   included in virtual search */
    char clanfilter[81];	/* selector for type of clan filtering:
				   (off, inclusive, strict) */

    /* *** for www server ***
       sprintf(argv_1,"/home/www/htdocs/Services/argos/repro/%s",argv[1]);
       prefix(argv_1,err_file_name);
       prefix(argv_1,out_file_name);
       sprintf(err_file_name,"%s.rep_err",err_file_name);
       sprintf(out_file_name,"%s.rep_out",out_file_name);
       stderr = fopen(err_file_name,"a+");
       stdout = fopen(out_file_name,"a+");
     */

    fprintf (stdout, 
	     "          **********************************************\n"
	     "          *                                            *\n"
	     "          *                  REPRO                     *\n"
	     "          *   Program for Detecting Distant Repeats    *\n"
	     "          *           in Protein Sequences             *\n"
	     "          *       (C) Jaap Heringa, EMBL, 1992         *\n"
	     "          *       (C) Jaap Heringa, NIMR, 1998         *\n"
	     "          *       (C) John Romein, VU, 2002, 2003      *\n"
	     "          *                                            *\n"
	     "          **********************************************\n\n");

    if (argc < 4
	|| (((act = toupper(argv[2][0])) == 'A' || act == 'L') && argc != 6
	    && argc != 7) || (act == 'C' && argc != 4 && argc != 5
			      && argc != 10 && argc != 11) || (act != 'C'
							       && act != 'A'
							       && act != 'L'
							       && argc != 5
							       && argc != 6))
	Showparam (argv[0]);


    /* determine the sequence length */
    strncpy (argv_1, argv[1], 199);

    strcpy (detfraglen, "MAXL");	/* default fragment length determination using selection 
					   of longest fragment */
    strcpy (virtualnodes, "ALL");	/* default selection of start-sites from which to infer 
					   virtual nodes */
    strcpy (clanfilter, "STRICT");	/* default selection of type of filtering on graph-clans */

    /* read the number of top script items */
    if (act == 'C')		/* read ali scripts from input file */
      {
	  if (argc == 5 || argc == 11)
	    {			/* check if the third argument represents a number */
		s = argv[3];
		if (*s == '-')
		    s++;
		for (; *s >= '0' && *s <= '9'; s++);
		if (*s == '\0')
		    sscanf (argv[3], "%ld", &topnum);
		if (topnum <= 0)
		    fatal("The number of top-alignments should be positive integer number");

		Verbose (argc, argv, &consconn, &minfraglen, detfraglen,
			 &virtualvar, virtualnodes, clanfilter);
		read_script_file (argv[4], subst_fn, topnum);
	    }
	  else			/* no topnum specified so take this number from scripts input file */
	    {
		Verbose (argc, argv, &consconn, &minfraglen, detfraglen,
			 &virtualvar, virtualnodes, clanfilter);
		read_script_file (argv[3], subst_fn, 0);
	    }
      }
    else
      {				/* no scripts file to read so calculate Needleman-Wunsch matrix */
	  paroffset = act == 'A' || act == 'L';	/* all parameters shifted one */
	  sscanf (argv[2 + paroffset], "%ld", &topnum);
	  if (topnum <= 0)
	      fatal("The number of top-alignments should be positive integer number");
	  strncpy (subst_fn, argv[3 + paroffset], 199);


	  /* check if substitute value or file is given */

	  if (argc == 6)
	    {			/* no argument given so take default PAM250 values */
		fprintf (stdout,
			 "substitution values by default set to PAM250 values%c",
			 '\n');
		mism_val = 1;
		Def_pam (V, alph);
	    }
	  else
	    {			/* check if the argument represents a number */
		s = argv[3 + paroffset];
		if (*s == '-')
		    s++;
		while (isdigit(*s))
		    s++;
		if (*s == '.')
		    s++;
		while (isdigit(*s))
		    s++;
		if (*s == '\0')
		    set_identity_matrix(argv[paroffset + 3], V);
		else
		  {		/* read a file containing alphabet and substitution weights */
		      Sp = ckopen (argv[3 + paroffset], "r");
		      mism_val = 0;
		      fprintf (stdout,
			       "...reading substitution values from file: %s.\n",
			       argv[3 + paroffset]);
		      fscanf (Sp, "%s", alph);
		      size = strlen (alph);
		      for (i = 0; i < size; i++)
			  for (j = 0; j <= i; j++)
			    {
				fscanf (Sp, "%f", &ms);
				V[alph[i]][alph[j]] = V[alph[j]][alph[i]] =
				    VMULT * ms;
			    }
		      fclose (Sp);
		  }
	    }
	  sscanf(argv[argc - 2], "%f", &q);
	  if (q < 0.0)
	      fatal("The gap-open penalty must be a nonnegative number");
	  Q = VMULT * q;
	  sscanf(argv[argc - 1], "%f", &r);
	  if (r < 0.0)
	      fatal("The gap-extend penalty must be a nonnegative number");
	  R = VMULT * r;

	  AliAddMatrix (A, M, V);

	  strcpy (detfraglen, "MAXL");	/* default determination using selection of longest fragment */
	  strcpy (virtualnodes, "ALL");	/* default selection of start-sites from which to infer 
					   virtual nodes */
	  strcpy (clanfilter, "STRICT");	/* default selection of type of filtering on graph-clans */

	  switch (act) {
	      case 'A' :
		      PrintIntScripts (mism_val, &script_h, argv_1, iscr_fn, subst_fn, Q, R);
		      printf("\nalignment scripts written into file: %s\n", iscr_fn);

		      PrintScripts (A, mism_val, &script_h, argv_1, scr_fn, subst_fn, Q, R);
		      printf("\ntop alignments written into file: %s\n", scr_fn);

		      printf ("end\n");	/* stop after performing alignment */
		      return;
	      case 'L' :
		      FileCedricScripts (A, M, &script_h, argv_1, iscr_fn);
		      printf("\nalignment pairwise residue scores written into file: %s\n", iscr_fn);
		      printf ("end\n");	/* stop after performing alignment */
		      return;
	    }
      }


    /* get output file names from sequence input file prefix */
    prefix (argv_1, scr_fn);
    sprintf (scr_fn, "%s.scr", scr_fn);

    ProcScript (A, &script_h, ckopen (scr_fn, "w"), argv_1, mism_val, subst_fn,
		Q, R, consconn, minfraglen, detfraglen, virtualvar,
		virtualnodes, clanfilter);

    fprintf (stdout, "\n");
    /* if (act != 'C')
       fprintf(stdout,"alignment scripts written into file: %s\n", iscr_fn);
     */
    fprintf (stdout, "alignments, graphs and repeats written into file: %s\n",
	     scr_fn);

    /* printf("similarities of fragments written into file: %s\n", sim_fn); */

    fprintf (stdout, "end.\n");
}


#if defined MULTI_THREADING

static void determine_nr_threads(void)
{
    char *env = getenv("NR_THREADS");

    if (env != 0)
	nr_threads = atoi(env);
#if defined __linux
    else {
	char buffer[256];
	FILE *file;

	nr_threads = 0;

	if ((file = fopen("/proc/cpuinfo", "r")) != 0) {
	    while (fgets(buffer, 256, file) != 0)
		if (strncmp(buffer, "processor", 9) == 0)
		    ++ nr_threads;

	    fclose(file);
	}
    }
#endif

    if (nr_threads <= 0)
	fatal("illegal number of threads");
}

#endif


void init(int *argc, char **argv)
{
    setbuf(stdout, 0);

#if defined TIMER
    timer_module_init();
#endif

#if defined MULTI_THREADING
    determine_nr_threads();
#endif

#if defined MULTI_PROCESSOR
    comm_init(argc, argv);
#endif

    read_sequence(argv[1]);
}


void end(void)
{
#if defined MULTI_PROCESSOR
    comm_end();
#endif
}


int main(int argc, char *argv[])
{
    init(&argc, argv);

#if defined MULTI_PROCESSOR
    if (node_id != master_node_id)
	AliAddMatrix(A, M, V);
    else
#endif
	do_work(argc, argv);

    end();
    return 0;
}


/*----------------------------------------------------------------------------*/

void Showparam (const char *name)
{
    fprintf (stdout,
	     "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
	     "You entered the program using an inproper parameter set:\n",
	     "there are three basic ways to run this program:\n", name,
	     "  seqfile a topnum [mismatch-val or Dayhoff-file] gap-i gap-e\n",
	     "or:\n", name, "  seqfile c [topnum] alignment-scripts-file\n",
	     "or:\n", name,
	     "  seqfile   topnum [mismatch-val or Dayhoff-file] gap-i gap-e\n",
	     "where:\n", "seqfile is the file containing the sequence\n",
	     "a (alignment) implies making alignments only (and automatically writing scripts to .iscr file)\n",
	     "c (continue) implies continuation with graph-based cluster analysis using a .iscr file\n",
	     "cv implies verbose continuation (more parameters)\n",
	     "no a or c implies a + c (so creating top alignments and analysis) \n",
	     "-  topnum is the number of non-overlapping top-scoring local alignments, generated using\n",
	     "   the 'a' option or analysed using 'c' option (topnum should be smaller or equal to that\n",
	     "   in previously produced .iscr file)\n",
	     "-  mismatch-val is negative real number applied to mismatching residues\n",
	     "-  Dayhoff-file is file containing residue substitution values\n",
	     "-  gap-i, gap-e, gap initiation and extension penalties (non-negative real value)\n",
	     "-  alignment-scripts-file is .iscr input file containing scripts from earlier run\n",
	     "\nFor time reasons it is advisable to first produce an .iscr file ('a' option) with\n",
	     "a high number of topscoring alignments (e.g. set topnum to 50) and then use this file\n",
	     "for further analysis using the 'c' (or 'cv') option. Producing the non-overlapping\n",
	     "top-scoring alignments is very slow compared to inferring the repeats during subsequent analysis\n",
	     "\n'l' instead of 'a' (alignment only option) yields alignment scripts file in T-Coffee library format\n");

    exit (0);
}

/*----------------------------------------------------------------------------*/

void Verbose (long argc, char *argv[], long *consconn, long *minfraglen, char *detfraglen, long *virtualvar, char *virtualnodes, char *clanfilter)
{
    char *s, str[81];

    if (toupper (argv[2][1]) == 'V' || argc > 5)	/* verbose mode, either from input line 
							   or requested from user */
      {
	  if (argc > 5)		/* consconn, minfraglen, detfraglen 
				   virtualvar, virtualnodes and clanfilter on input line */
	    {
		if (argc < 10)
		  {
		      fprintf (stderr,
			       "Error: less than 11 parameters on command line%c",
			       '\n');
		      Showparam (argv[0]);
		  }

		if (sscanf (argv[argc - 6], "%ld", consconn) <= 0
		    || *consconn <= 0)
		  {
		      fprintf (stderr,
			       "Error: value for consistentency required number of edges should be positive integer number%c",
			       '\n');
		      Showparam (argv[0]);
		  }
		;

		if (sscanf (argv[argc - 5], "%ld", minfraglen) <= 0
		    || *minfraglen <= 0)
		  {
		      fprintf (stderr,
			       "Error: value for minimal fragment length should be positive integer number%c",
			       '\n');
		      Showparam (argv[0]);
		  }
		;

		for (strcpy (detfraglen, argv[argc - 4]), s = detfraglen;
		     *s = toupper (*s); s++)
		    ;
		if (strcmp (detfraglen, "MEANL")
		    && strcmp (detfraglen, "MAXL"))
		  {
		      fprintf (stderr,
			       "inproper determination of fragment ends specified: %s\n",
			       detfraglen);
		      fprintf (stderr, "\n(options are MAXL or MEANL)\n");
		      Showparam (argv[0]);
		  }

		if (sscanf (argv[argc - 3], "%ld", virtualvar) <= 0
		    || *virtualvar < 0)
		  {
		      fprintf (stderr,
			       "Error: value for virtual fragment maximal length should be non-negative integer number%c",
			       '\n');
		      Showparam (argv[0]);
		  }
		;

		for (strcpy (virtualnodes, argv[argc - 2]), s = virtualnodes;
		     *s = toupper (*s); s++)
		    ;
		if (strcmp (virtualnodes, "N-TERMINAL")
		    && strcmp (virtualnodes, "ALL"))
		  {
		      fprintf (stderr,
			       "inproper selection of start-sites from which to infer virtual nodes: %s",
			       detfraglen);
		      fprintf (stderr, "%c(options are ALL or N-TERMINAL)\n",
			       '\n');
		      Showparam (argv[0]);
		  }

		for (strcpy (clanfilter, argv[argc - 1]), s = clanfilter;
		     *s = toupper (*s); s++)
		    ;
		if (strcmp (clanfilter, "OFF")
		    && strcmp (clanfilter, "INCLUSIVE")
		    && strcmp (clanfilter, "STRICT"))
		  {
		      fprintf (stderr,
			       "inproper selection of filtering new start-sites on graph-clans: %s",
			       clanfilter);
		      fprintf (stderr,
			       "%c(options are OFF, INCLUSIVE or STRICT)",
			       '\n');
		      Showparam (argv[0]);
		  }
	    }
	  else			/* go to user */
	    {
		/* no interactivity for use within server */

		/* 
		   fprintf(stderr,"no user around; -----> bye.%c",'\n');
		   exit(1);
		 */

		printf ("Verbose mode:\n");
		printf
		    ("Parameters can also be specified in order of following query\n");
		printf ("after other parameters on command line\n");

		Asks ("  enter minimal required graph-connection value [3]: ",
		      str);
		while (strlen (str) > 0
		       && (sscanf (str, "%ld", consconn) <= 0
			   || *consconn < 0))
		  {
		      printf
			  ("consistent graph-connection value should be positive integer\n");
		      Asks ("  enter maximal required graph-connection value [3]: ", str);
		  }
		if (strlen (str) == 0)
		    *consconn = 3;

		Asks ("  enter minimum value for repeating fragment length [1]: ", str);
		while (strlen (str) > 0
		       && (sscanf (str, "%ld", minfraglen) <= 0
			   || *minfraglen < 0))
		  {
		      printf
			  ("minimum fragment length should be positive integer value\n");
		      Asks ("  enter minimum value for repeating fragment length [1]: ", str);
		  }
		if (strlen (str) == 0)
		    *minfraglen = 1;

		Asks ("  enter indicator for determination of fragment ends [maxl]: ", str);
		for (s = str; *s = toupper (*s); s++)	/* convert to uppercase */
		    ;
		printf ("str = [%s], strlen = %d\n", str, strlen (str));
		while (strlen (str) && strcmp (str, "MEANL")
		       && strcmp (str, "MAXL"))
		  {
		      printf
			  ("inproper determination of fragment ends specified (enter \"maxl\" or \"meanl\")\n");
		      Asks ("  enter indicator for determination of fragment ends [maxl]: ", str);
		      for (s = str; *s = toupper (*s); s++)	/* convert to uppercase */
			  ;
		  }
		if (strlen (str))
		    strcpy (detfraglen, str);

		Asks ("  enter value for virtual fragment start-site variation [2]: ", str);
		while (strlen (str) > 0
		       && (sscanf (str, "%ld", virtualvar) <= 0
			   || *virtualvar < 0))
		  {
		      printf
			  ("virtual fragment variable length should be non-negative integer value\n");
		      Asks ("  enter value for virtual fragment start-site variation [2]: ", str);
		  }
		if (strlen (str) == 0)
		    *virtualvar = 2;

		Asks ("  enter selection of start-sites from which to infer virtual nodes [ALL]: ", str);
		for (s = str; *s = toupper (*s); s++)	/* convert to uppercase */
		    ;
		while (strlen (str) && strcmp (str, "N-TERMINAL")
		       && strcmp (str, "ALL"))
		  {
		      printf
			  ("inproper selection of start-sites (options are ALL or N-TERMINAL)\n");
		      Asks ("  enter selection of start-sites from which to infer virtual nodes [ALL]: ", str);
		      for (s = str; *s = toupper (*s); s++)	/* convert to uppercase */
			  ;
		  }
		if (strlen (str))
		    strcpy (virtualnodes, str);

		Asks ("  enter type of filter of new start-sites on graph-clans [STRICT]: ", str);
		for (s = str; *s = toupper (*s); s++)	/* convert to uppercase */
		    ;
		while (strlen (str) && strcmp (str, "OFF")
		       && strcmp (str, "INCLUSIVE") && strcmp (str, "STRICT"))
		  {
		      printf
			  ("inproper type of filtering (options are OFF, INCLUSIVE or STRICT)\n");
		      Asks ("  enter type of filter of new start-sites on graph-clans [STRICT]: ", str);
		      for (s = str; *s = toupper (*s); s++)	/* convert to uppercase */
			  ;
		  }
		if (strlen (str))
		    strcpy (clanfilter, str);
		;
	    }			/* end of else .... go to user */
      }
}

/*----------------------------------------------------------------------------*/

void Asks (char *s, char *str)
{
    printf (s);
    fgets (str, 80, stdin);
    str[strlen (str) - 1] = '\0';
    printf ("\n");
}

/*----------------------------------------------------------------------------*/


void Def_pam (long V[128][128], char alph[])
{
/* default is pam250  */
    int i, j, z;
    static const char let[24]  = { "ARNDCQEGHILKMFPSTWYVBZX" };
    static const long val[276] = {
       	2,
	-2, 6,
	0, 0, 2,
	0, -1, 2, 4,
	-2, -4, -4, -5, 12,
	0, 1, 1, 2, -5, 4,
	0, -1, 1, 3, -5, 2, 4,
	1, -3, 0, 1, -3, -1, 0, 5,
	-1, 2, 2, 1, -3, 3, 1, -2, 6,
	-1, -2, -2, -2, -2, -2, -2, -3, -2, 5,
	-2, -3, -3, -4, -6, -2, -3, -4, -2, 2, 6,
	-1, 3, 1, 0, -5, 1, 0, -2, 0, -2, -3, 5,
	-1, 0, -2, -3, -5, -1, -2, -3, -2, 2, 4, 0, 6,
	-4, -4, -4, -6, -4, -5, -5, -5, -2, 1, 2, -5, 0, 9,
	1, 0, -1, -1, -3, 0, -1, -1, 0, -2, -3, -1, -2, -5, 6,
	1, 0, 1, 0, 0, -1, 0, 1, -1, -1, -3, 0, -2, -3, 1, 2,
	1, -1, 0, 0, -2, -1, 0, 0, -1, 0, -2, 0, -1, -3, 0, 1, 3,
	-6, 2, -4, -7, -8, -5, -7, -7, -3, -5, -2, -3, -4, 0, -6, -2, -5, 17,
	-3, -4, -2, -4, 0, -4, -4, -5, 0, -1, -1, -4, -2, 7, -5, -3, -3, 0, 10,
	0, -2, -2, -2, -2, -2, -2, -1, -2, 4, 2, -2, 2, -1, -1, -1, 0, -6, -2, 4,
	0, -1, 2, 3, -4, 1, 2, 0, 1, -2, -3, 1, -2, -5, -1, 0, 0, -5, -3, -2, 2,
	0, 0, 1, 3, -5, 3, 3, -1, 2, -2, -3, 0, -2, -5, 0, 0, -1, -6, -4, -2, 2, 3,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    };

    for (i = 0, z = 0; i < 23; i++) {
	alph[i] = let[i];
	for (j = 0; j <= i; j++)
	    V[let[i]][let[j]] = V[let[j]][let[i]] = VMULT * val[z++];
    }

    printf("276 values loaded\n");
}

/*----------------------------------------------------------------------------*/

void read_script_file (char *scrfile, char *subst_fn, long topnum)
{
    FILE *fp, *fp2;
    char *s, s1[132], alph[129];
    long no_repeats;
    S_Script *ls, *cs;
    long *scr, i, j, size;
    float ms, score;
    float q, r;			/* gap penalties */

    fp = ckopen (scrfile, "r");

    fprintf (stdout, "...reading alignment scripts from file [%s]\n",
	     scrfile);
    fscanf (fp, "subst:%s", s1);
    if (!strcmp (s1, "default_PAM250"))
      {				/* default matrix should be taken */
	  mism_val = 1;
	  Def_pam (V, alph);
	  strcpy (subst_fn, "default PAM250");
      }
    else
      {				/* check if the argument represents a number */
	  s = s1;
	  if (*s == '-')
	      s++;
	  for (; *s >= '0' && *s <= '9'; s++);
	  if (*s == '.')
	      s++;
	  for (; *s >= '0' && *s <= '9'; s++);
	  if (*s == '\0')
	    {
		set_identity_matrix(s1, V);
		sprintf (subst_fn, "mismatch value: %6.2f", RVMULT * mism_val);
	    }
	  else
	    {			/* read a file containing alphabet and substitution weights */
		fp2 = ckopen (s1, "r");
		strcpy (subst_fn, s1);
		mism_val = 0;
		fprintf (stdout,
			 "...reading substitution values from file [%s]\n",
			 s1);
		fscanf (fp2, "%s", alph);
		size = strlen (alph);
		for (i = 0; i < size; i++)
		    for (j = 0; j <= i; j++)
		      {
			  fscanf (fp2, "%f", &ms);
			  V[alph[i]][alph[j]] = V[alph[j]][alph[i]] = VMULT * ms;
		      }
		fclose (fp2);
	    }
      }

    fscanf (fp, "%*s%f", &q);
    if (q < 0.0)
	fatal ("The gap-open penalty is nonnegative number");
    Q = VMULT * q;

    fscanf (fp, "%*s%f", &r);
    if (r < 0.0)
	fatal ("The gap-extend penalty is nonnegative number");
    R = VMULT * r;

    fprintf (stdout, "penalties: gap-open = %6.2f, gap-extent = %6.2f\n", q,
	     r);
    fscanf (fp, "%*s%ld", &no_repeats);
    if ((int) topnum)		/* topnum specified, so check if enough scripts */
      {
	  if ((int) topnum > (int) no_repeats)
	    {
		fprintf (stderr,
			 "only %ld scripts present in file [%s] instead of requested %ld scripts\n",
			 no_repeats, scrfile, topnum);
		exit (1);
	    }
	  no_repeats = topnum;
      }
    cs = &script_h;
    script_h.len = no_repeats;

    fprintf (stdout,
	     "...reading %ld highest scoring non-overlapping alignments\n",
	     script_h.len);
    for (i = 1; i <= no_repeats; i++, cs = ls)
      {
	  ls = (S_Script *) ckcalloc (1, sizeof (S_Script),
				      "Install_Script element");
	  cs->next = ls;
	  fscanf(fp, "%*s%*s%ld", &j);
	  if (i != j)
	      fatal("scripts out of sink");
	  fscanf(fp, "%*s%ld\n", &ls->yoffs);
	  fscanf(fp, "%*s%ld", &ls->x1);
	  fscanf(fp, "%*s%ld", &ls->y1);
	  fscanf(fp, "%*s%ld", &ls->x2);
	  fscanf(fp, "%*s%ld", &ls->y2);
	  fscanf(fp, "%*s%f", &score);
	  ls->score = VMULT * score;
	  fscanf(fp, "%*s%ld", &ls->size);
	  fscanf(fp, "%*s%ld", &ls->len);
	  scr = ckalloc(ls->len * sizeof (long), "Install_script script");
	  ls->lis = scr;
	  for (j = 0; j < ls->len; j++)
	      fscanf (fp, "%ld", scr + j);
      }

    script_h.score = ls->score;
    fclose (fp);
}


/*----------------------------------------------------------------------------*/


int		  *override;
score_type	  **good_scores;
score_vector_type *score_vector[128];
int		  nr_alignments = 0, nr_top_alignments_found = 0;

#if defined TIMER
struct timer      timer;
#endif


#if defined MULTI_THREADING
pthread_t	*threads;
pthread_mutex_t task_queue_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

#if defined MULTI_THREADING || defined MULTI_PROCESSOR
volatile int	nr_workers_active = 0;
#endif

#if defined MULTI_THREADING && !defined MULTI_PROCESSOR
pthread_cond_t	may_accept_work = PTHREAD_COND_INITIALIZER;
#endif

#if defined MULTI_PROCESSOR && defined MULTI_THREADING
pthread_mutex_t	comm_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
pthread_mutex_t long_message_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif


struct task {
    int			r, topnum, best_x, last_x;
    score_type		score;
    struct running_task *running_task;
    S_Script		*script;

#if defined MULTI_PROCESSOR
    struct task		*task_address;
#endif
} *tasks, **next_task, **task_queue;


struct running_task {
#if defined VERBOSE
    long long	start_time;
    struct task	old_task;

#if defined MULTI_PROCESSOR
    int		node_id;
#endif

#if defined MULTI_THREADING
    int		thread_id;
#endif
#elif !defined __GNUC__ && !defined __INTEL_COMPILER
    int		dummy_field;
#endif
};


unsigned   nr_tasks;
score_type best_unsorted_score = 0;
int	   pending_top_alignment_score = 0;
struct task *pending_top_alignment;

#if defined VERBOSE
score_type best_for_next_alignment = 0;
unsigned   max_to_go_this_alignment;
#endif


__inline__ static void allocate_good_scores(score_type **good_scores, struct task *task)
{
    int	y	= ((task->r - 1) & ~(BLOCK_SIZE - 1));
    int	max_y	= y + BLOCK_SIZE;
    int	length	= task->last_x - y;

    for (; y < max_y && -- length > 0; y ++)
	good_scores[y] = ckalloc(length * sizeof(score_type), "good_scores");
}


#if defined MULTI_PROCESSOR

struct message {
    void (*handler)(struct message *msg, int source);
    volatile bool  *received_reply;
    struct task    task, *slave_task;
    int		   thread_id;
};

#define UPCALL_TAG		1
#define LONG_MESSAGE_TAG	2


void poll(void)
{
    MPI_Status	   status;
    struct message msg;
    int		   msg_pending;

#if defined MULTI_THREADING
    pthread_mutex_lock(&comm_mutex);
#endif

    for (;;) {
	MPI_Iprobe(MPI_ANY_SOURCE, UPCALL_TAG, MPI_COMM_WORLD, &msg_pending, &status);

	if (!msg_pending)
	    break;

	MPI_Recv(&msg, sizeof msg, MPI_BYTE, MPI_ANY_SOURCE, UPCALL_TAG, MPI_COMM_WORLD, &status);

#if defined MULTI_THREADING
	pthread_mutex_unlock(&comm_mutex);
#endif
    
	msg.handler(&msg, status.MPI_SOURCE);

#if defined MULTI_THREADING
	pthread_mutex_lock(&comm_mutex);
#endif
    }

#if defined MULTI_THREADING
    pthread_mutex_unlock(&comm_mutex);
#endif
}


void poller(void)
{
    while (nr_top_alignments_found < topnum)
	poll();
}


void receive_good_scores(score_type **good_scores, struct task *task, int src)
{
    MPI_Status	status;
    int		y	= ((task->r - 1) & ~(BLOCK_SIZE - 1));
    int		max_y	= y + BLOCK_SIZE;
    int		length	= task->last_x - y;

    allocate_good_scores(good_scores, task);

#if defined MULTI_THREADING
    pthread_mutex_lock(&comm_mutex);
#endif

    for (; y < max_y && -- length > 0; y ++)
	MPI_Recv(good_scores[y], length, MPI_DATA_TYPE, src, LONG_MESSAGE_TAG, MPI_COMM_WORLD, &status);

#if defined MULTI_THREADING
    pthread_mutex_unlock(&comm_mutex);
#endif
}


void send_good_scores(score_type **good_scores, struct task *task, int dest)
{
    int y	= ((task->r - 1) & ~(BLOCK_SIZE - 1));
    int max_y	= y + BLOCK_SIZE;
    int length	= task->last_x - y;

#if defined MULTI_THREADING
    pthread_mutex_lock(&comm_mutex);
#endif

    for (; y < max_y && -- length > 0; y ++)
	MPI_Send(good_scores[y], length, MPI_DATA_TYPE, dest, LONG_MESSAGE_TAG, MPI_COMM_WORLD);

#if defined MULTI_THREADING
    pthread_mutex_unlock(&comm_mutex);
#endif
}

#endif


int compare_tasks(const void *p1, const void *p2)
{
    const struct task *t1  = * (const struct task **) p1;
    const struct task *t2  = * (const struct task **) p2;

#if defined MULTI_THREADING || defined MULTI_PROCESSOR
    int		      diff = (int) t2->running_task - (int) t1->running_task;

    if (diff != 0)
	return diff;
#endif

    return t2->score - t1->score;
}


__inline__ static void insert_task(struct task *task)
{
    if (task->score > best_unsorted_score)
	best_unsorted_score = task->score;

#if defined VERBOSE
    if (task->score > best_for_next_alignment) {
	struct task **t;

	best_for_next_alignment = task->score;

	for (t = next_task; t < task_queue + nr_tasks && (**t).score >= task->score; t ++)
	    ;

	max_to_go_this_alignment = BLOCK_SIZE * (t - next_task);
    }
#endif
}


__inline__ static struct task *get_task(void)
{
    if (next_task == task_queue + nr_tasks || (**next_task).score <= best_unsorted_score) {
	qsort(task_queue, nr_tasks, sizeof(struct task *), compare_tasks);
	best_unsorted_score = 0;
	next_task	    = task_queue;

#if defined MULTI_THREADING || defined MULTI_PROCESSOR
	while ((** next_task).running_task != 0)
	    next_task ++;
#endif
    }

    return *next_task ++;
}


#if BLOCK_SIZE == 8
#  if DATA_TYPE == SHORT_TYPE
#    define PZERO()		_mm_setzero_si128()
#    define PSETALL(X)		_mm_set1_epi16(X)
#    define PADD(X,Y)		_mm_adds_epi16(X,Y)
#    define PSUB(X,Y)		_mm_subs_epu16(X,Y)
#    define PMAX(X,Y)		_mm_max_epi16(X,Y)
#    define PAND(X,Y)		_mm_and_si128(X,Y)

const static union {
    short      i[8];
    block_type b;
} allow[7] = {{-1}, {-1, -1}, {-1, -1, -1}, {-1, -1, -1, -1}, {-1, -1, -1, -1, -1}, {-1, -1, -1, -1, -1, -1}, {-1, -1, -1, -1, -1, -1, -1}};

#  endif
#elif BLOCK_SIZE == 4
#  if DATA_TYPE == SHORT_TYPE
#    define PZERO()		_mm_setzero_si64()
#    define PSETALL(X)		_mm_set1_pi16(X)
#    define PADD(X,Y)		_m_paddsw(X,Y)
#    define PSUB(X,Y)		_m_psubusw(X,Y)
#    define PMAX(X,Y)		_m_pmaxsw(X,Y)
#    define PAND(X,Y)		_m_pand(X,Y)

const static union {
    short      i[4];
    block_type b;
} allow[3] = {{-1}, {-1, -1}, {-1, -1, -1}};

#  elif DATA_TYPE == FLOAT_TYPE
#    define PZERO()		_mm_setzero_ps()
#    define PSETALL(X)		_mm_set_ps1(X)
#    define PADD(X,Y)		_mm_add_ps(X,Y)
#    define PSUB(X,Y)		_mm_sub_ps(X,Y)
#    define PMAX(X,Y)		_mm_max_ps(X,Y)
#    define PAND(X,Y)		_mm_and_ps(X,Y)

const static union {
    int        i[4];
    block_type b;
} allow[3] = {{-1}, {-1, -1}, {-1, -1, -1}};

#  endif
#elif BLOCK_SIZE == 1
#    define PZERO()		0
#    define PSETALL(X)		(X)
#    define PADD(X,Y)		((X) + (Y))
#    define PSUB(X,Y)		((X) - (Y))
#    define PMAX(X,Y)		max(X,Y)
#    define PAND(X,Y)		((X) & (Y))
#endif

#if defined __OPTIMIZE__ && (DATA_TYPE == SHORT_TYPE && BLOCK_SIZE == 8 || DATA_TYPE == FLOAT_TYPE && BLOCK_SIZE == 4)
register block_type mpq __asm__("xmm6"), mpr __asm__("xmm7");
#elif defined __OPTIMIZE__ && DATA_TYPE == SHORT_TYPE && BLOCK_SIZE == 4
register block_type mpq __asm__("mm6"), mpr __asm__("mm7");
#else
block_type mpq, mpr;
#endif


__inline__ static void align_row(char A[], int last_x, score_vector_type *sv[128], block_type *m0, block_type *maxx, block_type *maxy, int x, int y)
{
    block_type	      t, vmaxx = maxx[y];
    score_vector_type *s = sv[A[y]];

    for (; x < last_x; x ++) {
	m0[x]   = PMAX(PZERO(), PADD(PSETALL(s[x]), PMAX(PMAX(vmaxx, maxy[x]), m0[x + 1])));
	t	= PSUB(m0[x + 1], mpq);
	vmaxx   = PSUB(PMAX(t, vmaxx), mpr);
	maxy[x] = PSUB(PMAX(t, maxy[x]), mpr);
    }

    maxx[y] = vmaxx;
}


__inline__ static void correct_left_border(block_type *m)
{
#if BLOCK_SIZE > 1
    int i;

    for (i = BLOCK_SIZE - 1; -- i >= 0;)
	m[i] = PAND(allow[i].b, m[i]);
#endif
}


__inline__ static void do_override(int last_x, int y, block_type *m0, int *ovv[])
{
    while (*ovv[y] < last_x)
	m0[*ovv[y] ++] = PZERO();
}


__inline__ static void check_best_score(score_type score, int x, int y1, struct task *task)
{
    if (score > task->score) {
	task->score  = score;
	task->best_x = x;
	task->r      = y1;
    }
}


__inline__ static void check_last_row(int y1, int begin, int end, struct task *task, score_type *p, score_type *gs)
{
    int x;

    for (x = end; -- x >= begin;)
	if (p[BLOCK_SIZE * x] == gs[x]) {
	    if (x + y1 + 1 > task->last_x)
		task->last_x = x + y1 + 1;

	    do
		if (p[BLOCK_SIZE * x] == gs[x])
		    check_best_score(gs[x], x, y1, task);
	    while (-- x >= begin);

	    return;
	}
}


__inline__ static void write_last_row(int y1, int begin, int end, struct task *task, score_type *p, score_type *gs)
{
    int x;

    for (x = end; -- x >= begin;) {
	gs[x] = p[BLOCK_SIZE * x];
	check_best_score(gs[x], x, y1, task);
    }
}


__inline__ static void do_row(char A[], int last_x, score_vector_type *sv[128], block_type *m0, block_type *maxx, block_type *maxy, int r, int x, int y, score_type **good_scores, struct task *task, int **ovv, int is_left, int is_bottom, int is_first)
{
    if (is_left)
	maxx[y] = m0[r + 1] = PZERO();

    align_row(A, last_x, sv, m0, maxx, maxy, x, y);

    if (is_left)
	correct_left_border(m0 + x);

    if (!is_first)
	do_override(last_x, y, m0, ovv);

    if (is_bottom) {
	int y1 = y + 1;

	(is_first ? write_last_row : check_last_row)(y1, is_left ? 0 : x - y1, last_x - y1, task, (score_type *) (m0 + y1) + y1 - r, good_scores[y]);
    }
}


__inline__ static void align_stripe(char A[], int M, score_vector_type *sv[128], int r, int x, block_type *m0, block_type *maxx, block_type *maxy, score_type **good_scores, struct task *task, int **ovv, int last_x, int is_left, int is_first)
{
    int y;

    if (last_x > x + LINE_SIZE)
	last_x = x + LINE_SIZE;

    memset(maxy + x, 0, (last_x - x) * sizeof(block_type));
    memset(m0 + x, 0, (last_x - x) * sizeof(block_type));

    for (y = 0; y < r - 1; y ++) {
	m0 -= 2;
	do_row(A, last_x, sv, m0, maxx, maxy, r, x, y, good_scores, task, ovv, is_left, 0, is_first);
    }

    for (; y < r + BLOCK_SIZE - 1 && y < M - 1; y ++) {
	m0 -= 2;
	do_row(A, last_x, sv, m0, maxx, maxy, r, x, y, good_scores, task, ovv, is_left, 1, is_first);
    }

#if 0 && defined MULTI_PROCESSOR
#if defined SSE || defined SSE2
    _mm_empty();
#endif

    if (node_id != master_node_id)
	poll();
#endif
}


void align_block(char A[], int M, score_vector_type *sv[128], score_type pq, score_type pr, block_type *m, block_type *maxx, block_type *maxy, score_type **good_scores, struct task *task, int *ovv[])
{
    int x, y, *ov, last_x, r = task->r;

    mpq = PSETALL(pq), mpr = PSETALL(pr);
    task->score = -1;

    if (task->topnum < 0) {
	align_stripe(A, M, sv, r, r, m, maxx, maxy, good_scores, task, ovv, M, 1, 1);

	for (x = r; (x += LINE_SIZE) < M;)
	    align_stripe(A, M, sv, r, x, m, maxx, maxy, good_scores, task, ovv, M, 0, 1);
    } else {
	last_x	     = task->last_x;
	task->last_x = -1;

	for (ov = override, y = 0; y < r + BLOCK_SIZE - 1 && y < M - 1; ov += topnum, y ++)
	    for (ovv[y] = ov; *ovv[y] < r; ++ ovv[y])
		;

	align_stripe(A, M, sv, r, r, m, maxx, maxy, good_scores, task, ovv, last_x, 1, 0);

	for (x = r; (x += LINE_SIZE) < last_x;)
	    align_stripe(A, M, sv, r, x, m, maxx, maxy, good_scores, task, ovv, last_x, 0, 0);
    }

#if defined SSE || defined SSE2
    _mm_empty();
#endif
}


void traceback_and_print(char A[], long M, score_vector_type *sv[128], long pq, long pr, score_type *m, long r, long x, long y, long score)
{
    char *str1 = ckalloc(2 * M, "string space"), *p1 = str1 + 2 * M - 1;
    char *str2 = ckalloc(2 * M, "string space"), *p2 = str2 + 2 * M - 1;
    int  gap;

    *p1 = *p2 = '\0';

    for (;;) {
	* -- p1 = A[r + x], * -- p2 = A[y];

	if ((score -= sv[A[y --]][r + x --]) == 0)
	    break;

	if (score != m[y * (M - r) + x])
	    for (gap = 1, score += pq + pr;; gap ++, score += pr) {
		if (x >= gap && score == m[y * (M - r) + x - gap]) {
		    x -= gap, p1 -= gap, p2 -= gap;
		    memcpy(p1, A + x + r + 1, gap);
		    memset(p2, '-', gap);
		    break;
		}

		if (y >= gap && score == m[(y - gap) * (M - r) + x]) {
		    y -= gap, p1 -= gap, p2 -= gap;
		    memcpy(p2, A + y + 1, gap);
		    memset(p1, '-', gap);
		    break;
		}
	    }
    }

    printf("%s\n%s\n", p1, p2);
    ckfree(str1);
    ckfree(str2);
}


__inline__ static void new_override(int x, int y, int *override)
{
    int *ptr1, *ptr2;

    for (ptr1 = override + topnum * y; *ptr1 < x; ++ ptr1)
	;

    assert(*ptr1 != x);

    for (ptr2 = ptr1; *ptr2 < M; ptr2 ++)
	;

    do
	ptr2[1] = ptr2[0];
    while (-- ptr2 >= ptr1);

    *ptr1 = x;
}


void set_overrides_from_script(S_Script *script, int *override)
{
    int	 i, j;
    int  x    = script->y1;
    int  y    = script->x1;
    long *scr = script->lis;

    for (i = 0; i < script->len; i ++)
	if (scr[i] > 0)
	    for (j = scr[i]; j >= 0; j --)
		new_override(x ++, y ++, override);
	else if (scr[i] < GAPYOFFS)
	    x += GAPYOFFS - scr[i] - 1;
	else
	    y += -scr[i] - 1;
}


#if defined MULTI_PROCESSOR

void broadcast_topalignment_handler(struct message *msg, int source)
{
    S_Script script;

#if defined MULTI_THREADING
    pthread_mutex_lock(&comm_mutex);
#endif

    ++ nr_top_alignments_found;
    MPI_Bcast(&script, sizeof script, MPI_BYTE, master_node_id, MPI_COMM_WORLD);
    script.lis = ckalloc(script.len * sizeof(long), "temp list");
    MPI_Bcast(script.lis, script.len, MPI_LONG, master_node_id, MPI_COMM_WORLD);
    set_overrides_from_script(&script, override);

#if defined MULTI_THREADING
    pthread_mutex_unlock(&comm_mutex);
#endif

    ckfree(script.lis);
}


void broadcast_topalignment(S_Script *script)
{
    struct message msg;
    int		   dest;

    msg.handler = broadcast_topalignment_handler;

#if defined MULTI_THREADING
    pthread_mutex_lock(&comm_mutex);
#endif

    for (dest = 0; dest < nr_nodes; dest ++)
	if (dest != node_id)
	    MPI_Send(&msg, sizeof msg, MPI_BYTE, dest, UPCALL_TAG, MPI_COMM_WORLD);

assert(script != 0);
assert(script->lis != 0);
    MPI_Bcast(script, sizeof *script, MPI_BYTE, master_node_id, MPI_COMM_WORLD);
    MPI_Bcast(script->lis, script->len, MPI_LONG, master_node_id, MPI_COMM_WORLD);

#if defined MULTI_THREADING
    pthread_mutex_unlock(&comm_mutex);
#endif
}

#endif


S_Script *traceback(char A[], long M, score_vector_type *sv[128], score_type pq, score_type pr, score_type *m, int r, int x, int y, score_type score, int *override)
{
    int gap, length, match_length, list_index;

    S_Script *script = ckalloc(sizeof *script, "top alignment");
    int	     *list   = ckalloc(M * sizeof(int), "temp list");
    int	     width   = (x + BLOCK_SIZE) & ~(BLOCK_SIZE - 1);

    script->yoffs    = r;
    script->x2	     = y;
    script->y2	     = r + x;
    script->score    = score;
    script->next     = 0;

    for (length = 1, match_length = 0, list_index = M;; ++ length) {
	/*new_override(x + r, y, override);*/

	if ((score -= sv[A[y --]][r + x --]) == 0)
	    break;

	if (score == m[y * width + x])
	    ++ match_length;
	else {
	    if (match_length > 0)
		list [-- list_index] = match_length;

	    for (gap = 0, score += pq;;) {
		++ gap, score += pr;

		if (x >= gap && score == m[y * width + x - gap]) {
		    x -= gap;
		    list [-- list_index] = GAPYOFFS - gap - 1;
		    break;
		}

		if (y >= gap && score == m[(y - gap) * width + x]) {
		    y -= gap;
		    list [-- list_index] = - gap - 1;
		    break;
		}
	    }

	    match_length = 0;
	    length += gap;
	}
    }

    if (match_length > 0)
	list [-- list_index] = match_length;

    script->x1	 = y + 1;
    script->y1	 = r + x + 1;
    script->size = length;
    script->len	 = M - list_index;
    script->lis	 = ckalloc(script->len * sizeof(long), "script");
    memcpy(script->lis, list + list_index, sizeof(long) * script->len);
    ckfree(list);
    return script;
}


#if !defined SSE

void compute_matrix(char A[], int M, score_vector_type *sv[128], score_type pq, score_type pr, score_type *matrix, score_type *maxy, int r, int best_x, score_type score, int *override)
{
    int		      x, y, width = best_x + 1;
    score_type	      maxx, t, *m = matrix;
    score_vector_type *s  = sv[A[0]] + r;
    int		      *ov = override;

    memset(maxy, 0, (best_x + 1) * sizeof *maxy);

    for (x = 0; x <= best_x; x ++)
	m[x] = max(0, s[x]);

    for (ov = override; (x = *ov - r) < best_x; ov ++)
	if (x >= 0)
	    m[x] = 0;

    for (y = 1; y < r; y ++) {
	m   += width;
	s    = sv[A[y]] + r;
	maxx = 0;
	m[0] = max(0, s[0]);

	for (x = 1; x <= best_x; x ++) {
	    t	    = m[x - width - 1];
	    m[x]    = max(0, s[x] + max(max(maxx, t), maxy[x]));
	    maxx    = max(t - pq, maxx) - pr;
	    maxy[x] = max(t - pq, maxy[x]) - pr;
	}

	for (override += topnum, ov = override; (x = *ov - r) < best_x; ov ++)
	    if (x >= 0)
		m[x] = 0;
    }

    assert(score == matrix[(r - 1) * width + best_x]);
}

#else

void compute_matrix(char A[], int M, score_vector_type *sv[128], score_type pq, score_type pr, score_type *matrix, score_type *maxy, int r, int best_x, score_type score, int *override)
{
    int		      x, y, width = (best_x + BLOCK_SIZE) / BLOCK_SIZE;
    score_type	      maxx, t;
    block_type	      mt, *m = (block_type *) matrix, *bmaxy = (block_type *) maxy;
    score_vector_type *s  = sv[A[0]] + r;
    int		      *ov = override;

    mpq = PSETALL(pq), mpr = PSETALL(pr);
    memset(maxy, 0, (best_x + 1) * sizeof *maxy);

    for (x = 0; x <= best_x; x ++)
	((score_type *) m)[x] = max(0, s[x]);

    for (ov = override; (x = *ov - r) < best_x; ov ++)
	if (x >= 0)
	    ((score_type *) m)[x] = 0;

    for (y = 1; y < r; y ++) {
	m   += width;
	s    = sv[A[y]] + r;
	maxx = 0;
	((score_type *) m)[0] = s[0];

	for (x = 1; x <= best_x; x ++) {
	    t	    = ((score_type *) m)[x - width * BLOCK_SIZE - 1];
	    ((score_type *) m)[x] = s[x] + max(max(maxx, t), maxy[x]);
	    maxx    = max(t - pq, maxx) - pr;
	}

	for (x = 0; x < width; x ++) {
	    m[x] = PMAX(m[x], PZERO());
	    mt = * (block_type *) ((score_type *) (m + x - width) - 1);
	    bmaxy[x] = PSUB(PMAX(PSUB(mt, mpq), bmaxy[x]), mpr);
	}

	for (override += topnum, ov = override; (x = *ov - r) < best_x; ov ++)
	    if (x >= 0)
		((score_type *) m)[x] = 0;
    }

    _mm_empty();
    assert(score == matrix[(r - 1) * width * BLOCK_SIZE + best_x]);
}

#endif


void topalignment(char A[], int M, score_vector_type *sv[128], score_type pq, score_type pr, struct task *task, int num, int *override)
{
#if defined TIMER
long long start_time = rdtsc();
#endif

    int	       width   = (task->best_x + BLOCK_SIZE) / BLOCK_SIZE;
    score_type *matrix = ckalloc(task->r * width * sizeof(block_type), "ali matrix data");
    score_type *maxy   = ckalloc(width * sizeof(block_type), "maxy array");;

    compute_matrix(A, M, sv, pq, pr, matrix, maxy, task->r, task->best_x, task->score, override);
    /*traceback_and_print(A, M, sv, pq, pr, matrix, task->r, task->best_x, task->r - 1, task->score);*/
    task->script = traceback(A, M, sv, pq, pr, matrix, task->r, task->best_x, task->r - 1, task->score, override);

    ckfree(matrix);
    ckfree(maxy);
#if defined TIMER
#if defined MULTI_PROCESSOR
printf("%d performed topalignment (r = %d, score = %d) in %6.3lf seconds\n", node_id, task->r, task->score, (rdtsc() - start_time) / Hz);
#else
printf("performed topalignment (r = %d, score = %d) in %6.3lf seconds\n", task->r, task->score, (rdtsc() - start_time) / Hz);
#endif
#endif
}


#if defined MULTI_THREADING || defined MULTI_PROCESSOR

__inline__ static void go_to_sleep(void)
{
#if defined MULTI_THREADING && defined MULTI_PROCESSOR
    pthread_mutex_unlock(&task_queue_mutex);
#endif

#if defined MULTI_PROCESSOR
    poll();
#endif

#if defined MULTI_THREADING && defined MULTI_PROCESSOR
    pthread_mutex_lock(&task_queue_mutex);
#endif

#if defined MULTI_THREADING && !defined MULTI_PROCESSOR
    pthread_cond_wait(&may_accept_work, &task_queue_mutex);
#endif
}

#endif


static void accept_top_alignment(struct task *task)
{
    S_Script	    *script	     = task->script;
    static S_Script *previous_script = &script_h;

    script_h.score	  = task->score;
    previous_script->next = script;
    previous_script	  = script;

#if defined MULTI_PROCESSOR
    broadcast_topalignment(script);
#endif

    set_overrides_from_script(script, override);
    ++ nr_top_alignments_found;
    printf("topalignment %d found, score = %g\n", nr_top_alignments_found, (float) task->score);
    pending_top_alignment_score = 0;

#if defined VERBOSE
    best_for_next_alignment = 0;
#endif

#if defined MULTI_THREADING && !defined MULTI_PROCESSOR
    pthread_cond_broadcast(&may_accept_work);
#endif
}


__inline__ static void reject_top_alignment(struct task *task)
{
    ckfree(task->script->lis);
    ckfree(task->script);
}


struct task *get_local_work(int source, int thread_id)
{
    struct task		*task;
    struct running_task *running_task;

#if defined MULTI_THREADING
    pthread_mutex_lock(&task_queue_mutex);
#endif

    for (;;) {
	if (nr_top_alignments_found == topnum) {
	    task = 0;
	    break;
	}
	
	task = get_task();

	if (task->score <= 0) {
	    task = 0; /* user asked for more alignments than available */
	    break;
	}

	if (task->score <= pending_top_alignment_score) {
	    -- next_task;

#if defined MULTI_THREADING || defined MULTI_PROCESSOR
	    if (nr_workers_active > 0)
		go_to_sleep();
	    else
#endif
		accept_top_alignment(pending_top_alignment);

	    continue;
	}

	if (task->topnum == nr_top_alignments_found)
	    pending_top_alignment_score = task->score;

#if defined MULTI_THREADING || defined MULTI_PROCESSOR
	++ nr_workers_active;
#endif
	running_task = ckalloc(sizeof *task->running_task, "running task");
	task->running_task = running_task;

#if defined VERBOSE
	task->running_task->old_task   = *task;
	task->running_task->start_time = rdtsc();

#if defined MULTI_PROCESSOR
	task->running_task->node_id    = source;
#endif

#if defined MULTI_THREADING
	task->running_task->thread_id  = thread_id;
#endif

	max_to_go_this_alignment -= BLOCK_SIZE;
#endif
	break;
    }

#if defined MULTI_THREADING
    pthread_mutex_unlock(&task_queue_mutex);
#endif

    return task;
}


#if defined MULTI_PROCESSOR

void fetch_good_scores_handler(struct message *msg, int source)
{
    send_good_scores(good_scores, &msg->task, source);
}


void work_reply_handler(struct message *msg, int source)
{
    if (msg->task.task_address != 0 && msg->task.topnum >= 0 && good_scores[msg->task.r - 1] == 0) {
	msg->handler = fetch_good_scores_handler;

#if defined MULTI_THREADING
	pthread_mutex_lock(&long_message_mutex);
	pthread_mutex_lock(&comm_mutex);
#endif

	MPI_Send(msg, sizeof *msg, MPI_BYTE, master_node_id, UPCALL_TAG, MPI_COMM_WORLD);

#if defined MULTI_THREADING
	pthread_mutex_unlock(&comm_mutex);
#endif

	receive_good_scores(good_scores, &msg->task, master_node_id);

#if defined MULTI_THREADING
	pthread_mutex_unlock(&long_message_mutex);
#endif
    }

    *msg->slave_task	 = msg->task;
    *msg->received_reply = true;
}


void work_request_handler(struct message *msg, int source)
{
    struct task *task = get_local_work(source, msg->thread_id);

    msg->handler = work_reply_handler;

    if (task != 0)
	msg->task = *task;
    else
	msg->task.task_address = 0;

#if defined MULTI_THREADING
    pthread_mutex_lock(&comm_mutex);
#endif

    MPI_Send(msg, sizeof *msg, MPI_BYTE, source, UPCALL_TAG, MPI_COMM_WORLD);

#if defined MULTI_THREADING
    pthread_mutex_unlock(&comm_mutex);
#endif
}


struct task *get_remote_work(int thread_id)
{
    MPI_Status	   status;
    struct message msg;
    pthread_cond_t receive_cond;
    volatile bool  received_reply = false;
    struct task	   *task = ckalloc(sizeof *task, "task");

    msg.handler		= work_request_handler;
    msg.received_reply	= &received_reply;
    msg.slave_task	= task;
    msg.thread_id	= thread_id;

#if defined MULTI_THREADING
    pthread_mutex_lock(&comm_mutex);
#endif

    MPI_Send(&msg, sizeof msg, MPI_BYTE, master_node_id, UPCALL_TAG, MPI_COMM_WORLD);

#if defined MULTI_THREADING
    pthread_mutex_unlock(&comm_mutex);
#endif

    while (!received_reply)
	poll();

    if (task->task_address == 0) {
	ckfree(task);
	task = 0;
    }

    return task;
}

#endif


__inline__ static struct task *get_work(int thread_id)
{
#if defined MULTI_PROCESSOR
    if (node_id != master_node_id)
	return get_remote_work(thread_id);
    else
	return get_local_work(master_node_id, thread_id);
#else
    return get_local_work(0, thread_id);
#endif
}


#if defined VERBOSE

__inline__ static void print_stats(struct task *task, struct running_task *running_task)
{
    long long	stop_time    = rdtsc();
    static int 	header_count = 0;

    if (-- header_count < 0) {
	header_count = 30;
	printf("CPU/T:      r last_x    max top |  score   bestsc time  aligns mxtogo tot.time\n");
    }

    printf( "%03d/%d:%7d %6d %6g %3d | %6g %6g %6.3lf %7d %6d %8.1lf\n",
#if defined MULTI_PROCESSOR
	    running_task->node_id,
#else
	    0,
#endif
#if defined MULTI_THREADING
	    running_task->thread_id,
#else
	    0,
#endif
	    running_task->old_task.r,
	    running_task->old_task.last_x,
	    (float) running_task->old_task.score,
	    running_task->old_task.topnum,
	    (float) task->score,
	    (float) best_for_next_alignment,
	    (stop_time - running_task->start_time) / Hz,
	    nr_alignments,
	    max_to_go_this_alignment,
	    (timer.total_time + stop_time) / Hz);
}

#endif


__inline__ static void finish_local_work(struct task *task)
{
    struct running_task *running_task = task->running_task;

#if defined MULTI_THREADING
    pthread_mutex_lock(&task_queue_mutex);
#endif

    nr_alignments += BLOCK_SIZE;

    if (task->topnum < nr_top_alignments_found) {
	task->topnum = nr_top_alignments_found;

#if defined VERBOSE
	print_stats(task, running_task);
#endif
    } else if (task->score >= pending_top_alignment_score)
	pending_top_alignment = task;

    insert_task(task);
    ckfree(running_task);
    task->running_task = 0;

#if defined MULTI_THREADING || defined MULTI_PROCESSOR
    -- nr_workers_active;
#endif

#if defined MULTI_THREADING
    pthread_mutex_unlock(&task_queue_mutex);
#endif
}


#if defined MULTI_PROCESSOR

void receive_script(struct task *task, int source)
{
    MPI_Status status;
    S_Script   *script;

#if defined MULTI_THREADING
    pthread_mutex_lock(&comm_mutex);
#endif

    task->script = script = ckalloc(sizeof *task->script, "script");
    MPI_Recv(script, sizeof *script, MPI_BYTE, source, LONG_MESSAGE_TAG, MPI_COMM_WORLD, &status);
    script->lis = ckalloc(script->len * sizeof(long), "script");
    MPI_Recv(script->lis, script->len, MPI_LONG, source, LONG_MESSAGE_TAG, MPI_COMM_WORLD, &status);

#if defined MULTI_THREADING
    pthread_mutex_unlock(&comm_mutex);
#endif
}


void send_script(struct task *task)
{
    S_Script *script = task->script;

    MPI_Send(script, sizeof *script, MPI_BYTE, master_node_id, LONG_MESSAGE_TAG, MPI_COMM_WORLD);
    MPI_Send(script->lis, script->len, MPI_LONG, master_node_id, LONG_MESSAGE_TAG, MPI_COMM_WORLD);
    ckfree(script->lis);
    ckfree(script);
}


static void finished_work_handler(struct message *msg, int source)
{
    struct task *task = &msg->task;

    if (task->topnum == nr_top_alignments_found)
	receive_script(task, source);
    else if (task->topnum < 0)
	receive_good_scores(good_scores, task, source);

    *task->task_address = *task;
    finish_local_work(task->task_address);
}


__inline__ static void finish_remote_work(struct task *task)
{
    struct message msg;

    msg.handler = finished_work_handler;
    msg.task	= *task;

#if defined MULTI_THREADING
    pthread_mutex_lock(&comm_mutex);
#endif

    MPI_Send(&msg, sizeof msg, MPI_BYTE, master_node_id, UPCALL_TAG, MPI_COMM_WORLD);

    if (task->topnum == nr_top_alignments_found)
	send_script(task);
    else if (task->topnum < 0)
	send_good_scores(good_scores, task, master_node_id);

#if defined MULTI_THREADING
    pthread_mutex_unlock(&comm_mutex);
#endif

    ckfree(task);
}

#endif


__inline__ static void finish_work(struct task *task)
{
#if defined MULTI_PROCESSOR
    if (node_id != master_node_id)
	finish_remote_work(task);
    else
#endif
	finish_local_work(task);
}


struct task original_tasks[2];
void slave(void *thread_id)
{
    struct task *task;
    block_type  *maxx = ckalloc(M * sizeof(block_type), "maxx array");
    block_type  *maxy = ckalloc(M * sizeof(block_type), "maxy array");
    int		**ovv = ckalloc((M - 1) * sizeof(int *), "override vector");
    block_type  *m    = ckalloc((3 * M + BLOCK_SIZE - 1) * sizeof(block_type), "ali matrix data");

    while ((task = get_work((int) thread_id)) != 0) {
	original_tasks[(int) thread_id] = *task;
	if (task->topnum < nr_top_alignments_found) {
#if BLOCK_SIZE > 1
	    task->r = ((task->r - 1) & ~(BLOCK_SIZE - 1)) + 1;
#endif
	    if (good_scores[task->r - 1] == 0)
		allocate_good_scores(good_scores, task);

	    align_block(A, M, score_vector, Q, R, m + 2 * M, maxx, maxy, good_scores, task, ovv);
	} else
	    topalignment(A, M, score_vector, Q, R, task, nr_top_alignments_found, override);

	finish_work(task);
    }

    ckfree(ovv);
    ckfree(m);
    ckfree(maxy);
    ckfree(maxx);
}


void *worker_thread(void *thread_id)
{
#if defined MULTI_PROCESSOR
    if (node_id == master_node_id && (int) thread_id == 0)
	poller();
    else
#endif
	slave(thread_id);

    return 0;
}


#if defined MULTI_THREADING

void start_threads(void)
{
    int i;

#if 0
    /* TODO */
    if (nr_threads > nr_tasks)
	nr_threads = nr_tasks;
#endif

    threads = ckalloc(nr_threads * sizeof(pthread_t), "threads");

    for (i = 0; i < nr_threads; i ++)
	if (pthread_create(threads + i, 0, worker_thread, (void *) i) != 0) {
	    perror("pthread_create");
	    exit(1);
	}
}


void stop_threads(void)
{
    int i;

    for (i = 0; i < nr_threads; i ++)
	if (pthread_join(threads[i], 0) != 0) {
	    perror("pthread_join");
	    exit(1);
	}

    ckfree(threads);
}


#endif


void AliAddMatrix(char A[], long M, long V[128][128])
{
    long i, x, width;

#if defined MULTI_PROCESSOR
    MPI_Bcast(&topnum, 1, MPI_LONG, master_node_id, MPI_COMM_WORLD);
    MPI_Bcast(&Q, 1, MPI_LONG, master_node_id, MPI_COMM_WORLD);
    MPI_Bcast(&R, 1, MPI_LONG, master_node_id, MPI_COMM_WORLD);
    MPI_Bcast(V, 128 * 128, MPI_LONG, master_node_id, MPI_COMM_WORLD);
#endif

    good_scores = ckcalloc(M - 1, sizeof(score_type *), "good_scores");
    override	= ckalloc((M - 1) * topnum * sizeof(int), "override positions");

    for (i = 0; i < M - 1; i ++) {
	/*good_scores[i] = ckalloc((M - 1 - i) * sizeof(score_type), "good_scores");*/
	override[topnum * i] = M;
    }

    for (i = 'A'; i <= 'Z'; i ++) {
	score_vector[i] = ckalloc(M * sizeof(score_vector_type), "score vector");

	for (x = 0; x < M; x ++)
	    score_vector[i][x] = V[i][A[x]];
    }

#if defined MULTI_PROCESSOR
    if (node_id == master_node_id) {
#endif
	width	    = (M + BLOCK_SIZE - 1) & ~BLOCK_SIZE;

	printf("...add matrix %ld x %ld using gap penalties (i, x) %6.2f and %6.2f\n", M, M, RVMULT * Q, RVMULT * R);

	nr_tasks = (M - 2) / BLOCK_SIZE + 1;
	tasks = ckalloc(nr_tasks * sizeof(struct task), "tasks");
	task_queue = next_task = ckalloc(nr_tasks * sizeof(struct task *), "tasks");

#if defined VERBOSE
	max_to_go_this_alignment = nr_tasks * BLOCK_SIZE;
#endif

	for (i = 0; i < nr_tasks; i ++) {
	    struct task *task = tasks + i;

	    task->r		= BLOCK_SIZE * i + 1;
	    task->topnum	= -1;
	    task->score		= MAX_SCORE;
	    task->last_x	= M;
	    task->running_task	= 0;

#if defined MULTI_PROCESSOR
	    task->task_address	= task;
#endif
	    task_queue[i]	= task;
	}

#if 1 /* SHUFFLE */
	for (i = 0; i < nr_tasks; i ++) {
	    int	    index     = nr_tasks * drand48();
	    struct task *task = task_queue[index];

	    task_queue[index] = task_queue[i];
	    task_queue[i]     = task;
	}
#endif

#if 0
    if (M == 34350) {
	task_queue[0] = tasks;
	tasks[0].r = 23364;
	tasks[0].topnum = 0;
	tasks[0].score = 16173;
	tasks[0].best_x = 8778;
    }
#endif

#if defined TIMER
	timer_init(&timer);
	timer_start(&timer);
#endif

#if defined MULTI_PROCESSOR
    }
#endif

#if !defined MULTI_THREADING
    worker_thread(0);
#else
    start_threads();
    stop_threads();
#endif

#if defined MULTI_PROCESSOR
    if (node_id == master_node_id) {
#endif

#if defined TIMER
	timer_stop(&timer);
	timer_print(&timer, "alignments timer");
#endif

	printf("nr alignments = %u\n", nr_alignments);
	script_h.len = nr_top_alignments_found;

#if defined MULTI_PROCESSOR
    }
#endif

    for (i = 0; i < M - 1; i ++)
	ckfree(good_scores[i]);

    for (i = 'A'; i <= 'Z'; i ++)
	ckfree(score_vector[i]);

    ckfree(tasks);
    ckfree(good_scores);
}


void ProcScript (char A[], S_Script * ptr, FILE * fp, char *seq_fn, long mismatch, char *comp_mat, long gapi, long gapx, long consconn, long minfraglen, char *detfraglen, long virtualvar, char *virtualnodes, char *clanfilter)
{
    S_Script *ls;
    long i = 0, scrdim, M, **aa = NULL;
    char *filter;

    fprintf (fp, "sequence read from file: %s\n", seq_fn);

    if (mismatch > 0)
	fprintf (fp,
		 "substitution values by default read from PAM-250 matrix\n");
    else if (!mismatch)
	fprintf (fp, "substitution values read from file: %s\n", comp_mat);
    else
	fprintf (fp, "match value = 1.00; mismatch value =%7.2f\n",
		 RVMULT * mismatch);
    fprintf (fp, "penalties: gap-open =%7.2f, gap-extend =%7.2f\n",
	     RVMULT * gapi, RVMULT * gapx);

    fprintf (fp, "script list: elements; %ld, lowest score: %6.2f\n",
	     ptr->len, RVMULT * ptr->score);

    aa = (long **) ckalloc ((M = strlen (A)) * sizeof (long *),
			    "multiple ali array: M long");
    aa[0] = (long *) ckcalloc (M * (scrdim = ptr->len + 1), sizeof (long),
			       "multiple ali data: (topno+1) x M");
    aa[0][0] = 1;


    for (i = 1; i < M; i++)
      {
	  aa[i] = aa[i - 1] + scrdim;
	  aa[i][0] = i + 1;
      }

    filter = (char *) ckalloc (scrdim, "filter");
    memset (filter, 1, scrdim);

    for (ls = ptr->next, i = 1; ls != NULL; ls = ls->next, i++)
      {
	  PrintAli (A, ls, i, fp);
	  StackAli (A, M, aa, ls, i, fp);
      }

    PrintStackAli (aa, A, M, scrdim, fp);

    fprintf (fp, "...evaluate stacked alignment for %ld alignments\n",
	     scrdim - 1);
    fprintf (stdout, "...evaluate stacked alignment for %ld alignments\n",
	     scrdim - 1);
    MakeGraphs (A, M, aa, ptr, scrdim, filter, fp, consconn, minfraglen,
		detfraglen, virtualvar, virtualnodes, clanfilter);

    if (aa != NULL)
      {
	  ckfree (aa[0]);
	  ckfree (aa);
      }
    if (filter != NULL)
	ckfree (filter);
}

/*----------------------------------------------------------------------------*/

void PrintScripts (char A[], long mismatch, S_Script * ptr, char *seq_fn, char *script_ofile, char *comp_mat, long gapi, long gapx)
{
    long i;
    FILE *ofp;
    char pre1[100];
    S_Script *ls;

    sprintf (script_ofile, "%s.scr", prefix (seq_fn, pre1));
    ofp = ckopen (script_ofile, "w");

    if (mismatch > 0)
      {
	  fprintf (ofp, "subst: default_PAM250\n");
      }
    else if (!mismatch)
      {
	  fprintf (ofp, "subst: %s\n", comp_mat);
      }
    else
      {
	  fprintf (ofp, "subst: %7.2f mismatch value, match value = 1.00\n",
		   RVMULT * mismatch);
      }

    fprintf (ofp, "gap-open: %7.2f\ngap-extend: %7.2f\n",
	     RVMULT * gapi, RVMULT * gapx);

    fprintf (ofp, "number-repeats: %ld\n", ptr->len);

    for (ls = ptr->next, i = 1; ls != NULL; ls = ls->next, i++)
      {
	  PrintAli (A, ls, i, ofp);
      }

    fclose (ofp);
}

/*----------------------------------------------------------------------------*/

void PrintIntScripts (long mismatch, S_Script * ptr, char *seq_fn, char *script_ofile, char *comp_mat, long gapi, long gapx)
{
    long i, j, *scr;
    FILE *ofp;
    char pre1[100];
    S_Script *ls;

    sprintf (script_ofile, "%s.iscr", prefix (seq_fn, pre1));
    ofp = ckopen (script_ofile, "w");

    if (mismatch > 0)
      {
	  fprintf (ofp, "subst: default_PAM250\n");
      }
    else if (!mismatch)
      {
	  /*
	     sprintf(script_ofile, "%s_%ld_%ld_%ld_%ld.iscr",
	     prefix(seq_fn, pre1), ptr->len, prefix(comp_mat, pre2), gapi, gapx);
	   */
	  fprintf (ofp, "subst: %s\n", comp_mat);
      }
    else
      {
	  fprintf (ofp, "subst: %7.2f mismatch value, match value = 1.00\n",
		   RVMULT * mismatch);
      }

    fprintf (ofp, "gap-open: %7.2f\ngap-extend: %7.2f\n",
	     RVMULT * gapi, RVMULT * gapx);

    fprintf (ofp, "number-repeats: %ld\n", ptr->len);

    for (ls = ptr->next, i = 1; ls != NULL; ls = ls->next, i++)
      {
	  fprintf (ofp, "----------\n");
	  fprintf (ofp, "repeat: %ld\n", i);
	  fprintf (ofp, "y-offset: %ld\n", ls->yoffs);
	  fprintf (ofp, "x1: %ld\n", ls->x1);
	  fprintf (ofp, "y1: %ld\n", ls->y1);
	  fprintf (ofp, "x2: %ld\n", ls->x2);
	  fprintf (ofp, "y2: %ld\n", ls->y2);
	  fprintf (ofp, "score: %.2f\n", RVMULT * ls->score);
	  fprintf (ofp, "length: %ld\n", ls->size);
	  fprintf (ofp, "script_length: %ld", ls->len);
	  for (j = 0, scr = ls->lis; j < ls->len; j++)
	      fprintf (ofp, "%s%6ld", (!(j % 10)) ? "\n" : " ", scr[j]);
	  fprintf (ofp, "%c", '\n');
      }

    fprintf (ofp, "CPU 0\n");
    fclose (ofp);
    fprintf (stdout, "...scripts in file [%s]\n", script_ofile);

}

/*----------------------------------------------------------------------------*/

void
FileCedricAli (S_Script * p, FILE * fp)
{
    long x, y, *scr, lim, i, j, score;

    if ((scr = p->lis) != NULL) {
	x = p->x1 + 1;
	y = p->y1 + 1;
	score = RVMULT * p->score;
	/* Cedric's format: #res1 #res2 (starting at 1) aliscr  consistency_number 0 (for future use) */
	fprintf (fp, "%6ld %6ld %12ld  1  0\n", x, y, score);

	for (lim = p->len, i = 0; i < lim; i++)
	    if (scr[i] > 0)	/* diagonal steps */
		for (j = 0; j < scr[i]; j++)
		    fprintf(fp, "%6ld %6ld %12ld  1  0\n", ++ x, ++ y, score);
	    else if (scr[i] < GAPYOFFS)	/* inserts in Y-sequence */
		fprintf(fp, "%6ld %6ld %12ld  1  0\n", ++ x, y += GAPYOFFS - scr[i], score);
	    else		/* inserts in X-sequence */
		fprintf(fp, "%6ld %6ld %12ld  1  0\n", x -= scr[i], ++ y, score);
    }
}

/*----------------------------------------------------------------------------*/

void FileCedricScripts (char A[], long M, S_Script * ptr, char *seq_fn, char *script_ofile)
{
    long i;
    FILE *ofp;
    char pre1[100];
    S_Script *ls;

    sprintf (script_ofile, "%s.locmat", prefix (seq_fn, pre1));
    ofp = ckopen (script_ofile, "w");
    fprintf (ofp, "1\n");	/* number of sequences */
    /* seqname seqlen aasequence */
    fprintf (ofp, "%s %ld ", prefix (seq_fn, pre1), M);
    for (i = 0; i < M; i++)
	fprintf (ofp, "%c", tolower (A[i]));
    fprintf (ofp, "\n");	/* sequence in lower case */
    fprintf (ofp, "#0 0\n");	/* line to indicate what sequence against what; here sequence(0) against itself */

    for (ls = ptr->next; ls != NULL; ls = ls->next)
	FileCedricAli (ls, ofp);

    fclose (ofp);
    fprintf (stdout, "...Position-wise local alignment scores in file [%s]\n",
	     script_ofile);

}

/*----------------------------------------------------------------------------*/

void
PrintAli (char A[], S_Script * p, long num, FILE * fp)
  /*
     char A[];        sequence   
     S_Script *p;     ptr to alignment script entry   
     long num;
     FILE *fp;
   */
{
    long dim, x, y, *scr, lim, lim2, i, j, k, l, t, ind, loop;
    char *a, *b, *c;

    dim = p->x2 - p->x1 + 1 + p->y2 - p->y1 + 1;

    a = ckalloc (dim, "Printfrag a");
    b = ckalloc (dim, "Printfrag b");
    c = ckalloc (dim, "Printfrag c");

    if ((scr = p->lis) != NULL)
      {
	  x = p->x1;
	  y = p->y1;
	  ind = 0;
	  a[ind] = A[x];
	  b[ind] = A[y];
	  c[ind] = (a[ind] == b[ind]) ? '|' : ' ';
	  for (lim = p->len, i = 0; i < lim; i++)
	    {
		if ((lim2 = scr[i]) > 0)	/* diagonal steps */
		    for (j = 0; j < lim2; j++)
		      {
			  a[++ind] = A[++x];
			  b[ind] = A[++y];
			  c[ind] = (a[ind] == b[ind]) ? '|' : ' ';
		      }
		else if (lim2 < GAPYOFFS)	/* inserts in Y-sequence */
		  {
		      lim2 = GAPYOFFS - lim2;
		      for (j = 1; j < lim2; j++)
			{
			    a[++ind] = '-';
			    b[ind] = A[++y];
			    c[ind] = ' ';
			}
		      a[++ind] = A[++x];
		      b[ind] = A[++y];
		      c[ind] = (a[ind] == b[ind]) ? '|' : ' ';
		  }
		else
		  {		/* inserts in X-sequence */
		      lim2 = -lim2;
		      for (j = 1; j < lim2; j++)
			{
			    a[++ind] = A[++x];
			    b[ind] = '-';
			    c[ind] = ' ';
			}
		      a[++ind] = A[++x];
		      b[ind] = A[++y];
		      c[ind] = (a[ind] == b[ind]) ? '|' : ' ';
		  }
	    }

	  /*
	     a[++ind] = b[ind] = c[ind] = '\0';
	     fprintf(fp, "->%s\n->%s\n->%s\n\n", a, c, b);
	   */
	  ++ind;

	  fprintf (fp,
		   "repeat %4ld   length = %ld   score = %10.2f   mean score = %10.2f\n",
		   num,
		   ind,
		   RVMULT * (p->score),
		   (ind > 0) ? RVMULT * p->score / ind : 0);

	  for (loop = 0, j = 0, k = p->x1 + 1, l = p->y1 + 1; j < ind;)
	    {
		loop = (ind - j > PWLINLEN) ? j + PWLINLEN : ind;
		fprintf (fp, "\n%5ld ", k);
		t = j;
		for (; j < loop; j++)
		  {
		      if (!(j % 10))
			  fprintf (fp, " %c", a[j]);
		      else
			  fprintf (fp, "%c", a[j]);
		      if (a[j] != '-')
			  k++;
		  }
		fprintf (fp, "%5ld\n      ", k - 1);
		for (j = t; j < loop; j++)
		  {
		      if (!(j % 10))
			  fprintf (fp, " %c", c[j]);
		      else
			  fprintf (fp, "%c", c[j]);
		  }
		fprintf (fp, "\n%5ld ", l);
		for (j = t; j < loop; j++)
		  {
		      if (!(j % 10))
			  fprintf (fp, " %c", b[j]);
		      else
			  fprintf (fp, "%c", b[j]);
		      if (b[j] != '-')
			  l++;
		  }

		fprintf (fp, "%5ld\n", l - 1);

	    }
	  fprintf (fp, "\n%s%s\n",
		   "----------------------------------------",
		   "----------------------------------------");
      }

    ckfree (a);
    ckfree (b);
    ckfree (c);
}

/*------------------------------------------------------------------------*/

void
StackAli (char A[], long M, long **ali, S_Script * p, long num, FILE * fp)
{
    long x, y, *scr, lim, lim2, i, j, ind; 

    if ((scr = p->lis) != NULL)
      {
	  x = p->x1;
	  y = p->y1;
	  ind = 0;
	  ali[x][num] = y + 1;
	  ali[y][num] = x + 1;

	  for (lim = p->len, i = 0; i < lim; i++)
	    {
		if ((lim2 = scr[i]) > 0)	/* diagonal steps */
		    for (j = 0; j < lim2; j++)
		      {
			  ali[++x][num] = ++y + 1;
			  ali[y][num] = x + 1;
		      }
		else if (lim2 < GAPYOFFS)	/* inserts in Y-sequence */
		  {
		      ali[x][num] = -ali[x][num];
		      lim2 = GAPYOFFS - lim2;
		      for (j = 1; j < lim2; j++)
			{
			    ali[++y][num] = -M - 1;
			}
		      ali[++x][num] = -(++y + 1);
		      ali[y][num] = x + 1;
		  }
		else
		  {		/* inserts in X-sequence */
		      ali[y][num] = -ali[y][num];
		      lim2 = -lim2;
		      for (j = 1; j < lim2; j++)
			{
			    ali[++x][num] = -M - 1;
			}
		      ali[++y][num] = -(++x + 1);
		      ali[x][num] = y + 1;
		  }
	    }
      }
}

/*------------------------------------------------------------------------*/

void
PrintStackAli (long **ali, char A[], long M, long dim, FILE * fp)
{
    long i, j, k, t;

    fprintf (fp, "Stacked alignments:\n");

    for (i = SALINLEN; i < M; i += SALINLEN)
      {
	  fprintf (fp, "\n      ");

	  for (j = i - SALINLEN; j < i; j++)
	    {
		if (!((j + 1) % 10))
		    fprintf (fp, "%-10ld", j + 1);
		else if (!(j % SALINLEN))
		    fprintf (fp, "%-9ld", j + 1);
	    }

	  for (k = 0; k < dim; k++)
	    {
		fprintf (fp, "\n%4ld  ", k);

		for (j = i - SALINLEN; j < i; j++)
		  {
		      if ((t = ali[j][k]) > 0)
			  fprintf (fp, "%c", A[t - 1]);
		      else if (t < 0)
			{
			    if (t < -M)
				fprintf (fp, "-");
			    else
				fprintf (fp, "%c", tolower (A[-(t + 1)]));
			}
		      else
			  fprintf (fp, ".");
		  }
	    }
	  fprintf (fp, "%c", '\n');
      }

    for (i -= SALINLEN; i < M; i += SALINLEN)
      {
	  fprintf (fp, "\n      ");

	  for (j = i; j < M; j++)
	      if (!((j + 1) % 10))
		  fprintf (fp, "%-10ld", j + 1);
	      else if (!(j % SALINLEN))
		  fprintf (fp, "%-9ld", j + 1);

	  for (k = 0; k < dim; k++)
	    {
		fprintf (fp, "\n%4ld  ", k);

		for (j = i; j < M; j++)
		  {
		      if ((t = ali[j][k]) > 0)
			  fprintf (fp, "%c", A[t - 1]);
		      else if (t < 0)
			{
			    if (t < -M)
				fprintf (fp, "-");
			    else
				fprintf (fp, "%c", tolower (A[-(t + 1)]));
			}
		      else
			  fprintf (fp, ".");
		  }
	    }
      }
    fprintf (fp, "\n\n");
}

/*--------------------------------------------------------------------------*/

void
RecRealStart2 (long *outstart, long i, long ii,
	       long alinum, long *allstartnum, long ***allstart,
	       char *filter, long dim, long M, long **ali, long *wght,
	       long *check)
{
    long j, k, m, n, t;
    /*
       static long print;
       long tttt; 
       print++;
       for (tttt =0; tttt < print; tttt++)
       printf("X");
       printf(" %ld\n", ii);
     */

    for (j = 1; j < dim; j++)
	if (filter[j] && j != alinum)
	  {
	      for (k = 0; k < 2; k++)
		{
		    for (m = 0, n = allstartnum[j]; m < n; m++)
		      {
			  if ((t = allstart[j][k][m]) == ii)
			    {
				if (k == 0)
				    t = allstart[j][1][m];
				else
				    t = allstart[j][0][m];

				if (!(check[t]))
				  {
				      check[t]++;
				      outstart[i]++;

				      RecRealStart2 (outstart, i, t, j,
						     allstartnum, allstart,
						     filter, dim, M, ali,
						     wght, check);
				  }
				/*
				   else
				   {
				   check[t]++;
				   outstart[i]++;
				   }
				 */
			    }
		      }
		}
	  }
    /*
       print--;
     */
}

/*-------------------------------------------------------------*/

void
RecRealStart (long **pali, long index1, long index2, long i, long ii,
	      long alinum, long **start, char *filter,
	      long dim, long M, long **ali, long *wght, long *check,
	      long virtualvar)
{
    long j, t, d1, d2;
    /*
       long tttt; 
       print++;
       for (tttt =0; tttt < print; tttt++)
       printf("X");
       printf(" %ld\n", ii);
     */

    for (j = 1; j < dim; j++)
      {
	  if (filter[j] && j != alinum)
	    {
		if (((d1 = _Abs (start[j][0] - ii)) <= virtualvar
		     || (d2 = _Abs (start[j][1] - ii)) <= virtualvar))
		  {
		      if ((t = ali[ii][j]) && t > -M)	/* no blank or gap and close after
							   start */
			{
			    if (t < 0)
				t = -(t + 1);
			    else
				t--;

			    if (!(check[t]))
			      {
				  check[t]++;
				  pali[i][index1]++;
				  pali[i][index2] += wght[j];

				  RecRealStart (pali, index1, index2, i, t, j,
						start, filter, dim, M, ali,
						wght, check, virtualvar);
			      }
			    else
			      {
				  pali[i][index1]++;
				  pali[i][index2] += wght[j];
			      }
			}
		      else if (!t)	/* blank */
			{
			    if (d1 <= virtualvar && ii < start[j][0])
				/* close before start[j][0] */
			      {
				  if ((t = start[j][0] - d1) >= 0)
				    {
					if (!(check[t]))
					  {
					      check[t]++;
					      pali[i][index1]++;
					      pali[i][index2] += wght[j];

					      RecRealStart (pali, index1,
							    index2, i, t, j,
							    start, filter,
							    dim, M, ali, wght,
							    check,
							    virtualvar);
					  }
					else
					  {
					      pali[i][index1]++;
					      pali[i][index2] += wght[j];
					  }
				    }
			      }
			    else /* close before start[j][1] */ if (ii > start[j][3])	/* close after end of N-terminal fragment */
			      {
				  if ((t = start[j][1] - d2) >= 0)
				    {
					if (!(check[t]))
					  {
					      check[t]++;
					      pali[i][index1]++;
					      pali[i][index2] += wght[j];

					      RecRealStart (pali, index1,
							    index2, i, t, j,
							    start, filter,
							    dim, M, ali, wght,
							    check,
							    virtualvar);
					  }
					else
					  {
					      pali[i][index1]++;
					      pali[i][index2] += wght[j];
					  }
				    }
			      }
			}
		  }
	    }
      }
    /*
       print--;
     */
}

/*-------------------------------------------------------------*/

void
RecStart (long **pali, long index1, long index2, long i, long ii,
	  char *filter, long dim, long M, long **ali, long *wght, long *check)
{
    long j, t2;

    for (j = 1; j < dim; j++)
	if (filter[j] && (t2 = ali[ii][j]) && t2 >= -M)	/* no blank or gap */
	  {
	      if (t2 < 0)
		  t2 = -(t2 + 1);
	      else
		  t2--;

	      if (!(check[t2]))	/* new startpoint */
		{
		    check[t2]++;
		    pali[i][index1]++;
		    pali[i][index2] += wght[j];

		    RecStart (pali, index1, index2, i, t2, filter, dim, M,
			      ali, wght, check);
		}
	      else
		{
		    pali[i][index1]++;
		    pali[i][index2] += wght[j];
		}
	  }
}

/***
void RecStart(long **pali, long index1, long index2, long i,
	      char *filter, long dim, long M, long **ali, long *wght)
{
  long j, t2;

  for (j = 1; j < dim; j++)
    if (filter[j] && (t2 = ali[i][j]) && t2 >= -M)   
      {
      if (t2 < 0) t2 = -(t2+1);
      else t2--;
  
      if (!(pali[t2][index1]))  
	{
	pali[t2][index1]++;
	pali[t2][index2] += wght[j];

	RecStart(pali, index1, index2, t2, filter, dim, M, ali, wght);
	}
      else
	{
	pali[t2][index1]++;
	pali[t2][index2] += wght[j];
	}
      }
}
***/
/*-------------------------------------------------------------*/

void
StackAll (long *workcount, long *workwght, long i,
	  char *filter, long dim, long M, long **ali, long *wght)
{
    long j, t2;
    /*
       static long print;
       long t; 

       print++;
       for (t =0; t < print; t++)
       printf("X");
       printf(" %ld\n", i);
     */

    for (j = 1; j < dim; j++)
	if (filter[j] && (t2 = ali[i][j]) && t2 >= -M)	/* no blank or gap */
	  {
	      if (t2 < 0)
		  t2 = -(t2 + 1);
	      else
		  t2--;

	      if (!(workcount[t2]))	/* new startpoint */
		{
		    workcount[t2]++;
		    workwght[t2] += wght[j];

		    StackAll (workcount, workwght, t2, filter, dim, M, ali,
			      wght);
		}
	      else
		{
		    workcount[t2]++;
		    workwght[t2] += wght[j];
		}
	  }
    /*
       print--;
     */
}

/*-------------------------------------------------------------*/

void
MakeGraphs (char A[], long M, long **ali, S_Script * p,
	    long dim, char *filter, FILE * ofp,
	    long consconn, long minfraglen,
	    char *detfraglen, long virtualvar, char *virtualnodes,
	    char *clanfilter)
{
    S_Script *ls;
    long **scr, i, j, k, *wght, *lenwght, t, t2, t3, t4,
	*start, *workcount, *workwght,
	*allstartnum, ***allstart, max, maxcount,
	maxloc, **addstart, addstartnum, loopcount, minlength, graphnum,
	graphamount, **startgraph =
	NULL, *locgraph, newstart, totnewstart, *graphclan, newclan,
	totnewclan, **subgraph, ***adjmat, subgraphnum, *oldallstartnum,
	**frags, **initstartgraph, *initallstartnum, initgraphamount,
	***initadjmat, fragcount, fragstop;
    S_Subgraph *lsubgraph;
    char graphtitle[30];

    /* alloc table for start and stop sites for each alignment script */
    alloc_2d (scr, long, dim, 4,
	      "ptrs for 2-d array of scripts start- and endpoints");

    wght = (long *) ckalloc (dim * sizeof (long), "ali wght array");

    lenwght = (long *) ckalloc (dim * sizeof (long), "ali len wght array");

    graphclan = (long *) ckalloc (dim * sizeof (long), "graphclan array");

    allstartnum = (long *) ckalloc (dim * sizeof (long), "allstartnum array");

    oldallstartnum = (long *) ckalloc (dim * sizeof (long),
				       "oldallstartnum array");

    alloc_3d (allstart, long, dim, 2, ALLSTARTLIM, "allstart 3d array");
    /*
       allstart        = (long ***) ckalloc(dim*sizeof(long **), 
       "allstart array 3-d ptr");
       allstart[0]     = (long **) ckalloc(2*dim*sizeof(long *), 
       "allstart array 2-d ptr");
       allstart[0][0]  = (long *) ckalloc(2*dim*ALLSTARTLIM*sizeof(long), 
       "allstart array 1-d ptr");
       for (i = 0, help = allstart[0][0]; i < dim; i++)
       {
       if (i)
       allstart[i] = allstart[i-1] + 2;
       for (j = 0; j < 2; j++)
       {
       allstart[i][j] = help;
       help += ALLSTARTLIM; 
       }
       }
     */

    alloc_2d (addstart, long, 3, TEMPSTARTLIM, "2-d array of infered starts");

    workcount = (long *) ckalloc (M * sizeof (long),
				  "Stack all alis's work counts array");

    workwght = (long *) ckcalloc (M, sizeof (long),
				  "Stack all alis's work wghts array");

    start = (long *) ckcalloc (M, sizeof (long), "real starts array");

    for (ls = p->next, i = 1; ls != NULL; ls = ls->next, i++)
      {
	  scr[i][0] = ls->x1, scr[i][1] = ls->y1,
	      scr[i][2] = ls->x2, scr[i][3] = ls->y2;

	  wght[i] = (ls->size > 0) ? ls->score / ls->size : 0;
	  lenwght[i] = ls->score;

	  allstart[i][0][0] = ls->x1;
	  allstart[i][1][0] = ls->y1;
	  allstartnum[i] = 1;
      }

    for (totnewstart = totnewclan = 1, loopcount = 0;
	 totnewstart || totnewclan; loopcount++)
      {
	  fprintf (stdout,
		   "========= graph cluster iteration %ld ==========\n",
		   loopcount + 1);
	  totnewstart = 0;
	  totnewclan = 0;
	  memset (start, 0, M * sizeof (long));	/* zero starts  output array array */
	  memset (workcount, 0, M * sizeof (long));	/* zero work counts array */

	  for (i = 1, graphamount = 0; i < dim; i++)
	      if (filter[i])
		  for (j = 0; j < 2; j++)	/* all start pairs */
		      for (k = 0, t = allstartnum[i]; k < t; k++)
			{
			    if (!(workcount[(t2 = allstart[i][j][k])]))	/* only one time per startsite */
			      {
				  workcount[t2]++;	/* set current start position to 1 */
				  start[t2]++;
				  RecRealStart2 (start, t2, t2,
						 0, allstartnum, allstart,
						 filter, dim, M, ali, wght,
						 workcount);
				  graphamount++;
			      }
			}


	  for (i = 1; i < dim; i++)	/* copy allstartnum to oldallstartnum to let
					   graphs start with old number of members */
	      oldallstartnum[i] = allstartnum[i];

	  for (i = 0, maxcount = 0; i < M; i++)
	    {
		if (start[i] > maxcount)
		    maxcount = start[i], maxloc = i;
	    }

	  /* now we know size of biggest graph, make array
	     startgraph[graphamount][maxcount] to memorize different connected graphs 
	     startgraph[i][0] contains number of nodes in graph */
	  if (startgraph != NULL)
	    {
		free_2d (startgraph);
	    }

	  alloc_2d (startgraph, long, graphamount, (maxcount + 1),
		    "graphs array");

	  if (loopcount == 0)	/* initial graph set-up */
	      alloc_2d (initstartgraph, long, graphamount, (maxcount + 1),
			"graphs array");

	  /* now find all different startgraphs when there are new startsites */
	  for (graphnum = 0; graphnum < graphamount; graphnum++)
	      /* graphs loop */
	    {
		memset (graphclan, 0, dim * sizeof (long));

		for (i = 0, maxcount = 0; i < M; i++)
		    if (start[i] > maxcount)
			maxcount = start[i], maxloc = i;
		start[maxloc] = 0;	/* will not be inspected anymore */

		/* now find nodes of this graph */
		memset (workcount, 0, M * sizeof (long));	/* zero work counts array */
		workcount[maxloc]++;	/* set current start position to 1 */
		/* no problem now to doubly use workcount array during recursion */
		RecRealStart2 (workcount, maxloc, maxloc,
			       0, oldallstartnum, allstart, filter,
			       dim, M, ali, wght, workcount);

		/* copy graph into startgraph array and print */
		fprintf (stdout, "graph %ld: %ld start-sites", graphnum + 1,
			 maxcount);
		startgraph[graphnum][0] = maxcount;
		for (i = 0, j = 0, max = 0; i < M; i++)
		    if (workcount[i])
		      {
			  startgraph[graphnum][++j] = i;
			  fprintf (stdout, "%s%4ld",
				   (!((max++) % 10)) ? "\n " : " ", i + 1);
		      }
		fprintf (stdout, "\n");
		assert(maxcount == max); /*fatalf ("internal error in graph %ld.\n", graphnum + 1); */

		if (loopcount == 0)	/* initgraphs only in first loop */
		  {
		      initgraphamount = graphamount;
		      initstartgraph[graphnum][0] = maxcount;
		      for (i = j = 0; i < M; i++)
			  if (workcount[i])
			      initstartgraph[graphnum][++j] = i;
		  }

		locgraph = &(startgraph[graphnum][1]);	/* locgraph is help ptr */

    /*************************** determine shortest repeat length
    if ((maxcount = startgraph[graphnum][0]) >= 2)
      {
      for (i = 1, 
	   minlength = locgraph[1] - locgraph[0]; 
	   i < maxcount; i++)
	if (minlength > 
	     (t = locgraph[i] - locgraph[i-1]))
	  minlength = t;
      }
    else
      minlength = 0;
    **********************/

		/* check if there are scripts that start just past two start-sites
		   in current start-graph. if so, copy this new pair into start-pairs
		   array so script becomes "contributing member" of this graph */
		addstartnum = 0;	/* empty addstarts array */

		/* find close virtual start-sites */
		if (loopcount == 0)
		    CopyCloseStarts (startgraph, graphnum, allstart,
				     allstartnum, addstart, &addstartnum,
				     filter, dim, -1, scr, virtualvar,
				     virtualnodes);
		else
		    CopyCloseStarts (startgraph, graphnum, allstart,
				     allstartnum, addstart, &addstartnum,
				     filter, dim, 1, scr, virtualvar,
				     virtualnodes);

		/* now try to find infered start-sites, i.e. sites that are in the
		   middle of a repeat and  hit at least one of the determined start-sites,
		   or sites that are preceding a repeat (so repeat must be made longer
		   in N-terminal direction) and hit at least one determined start-site
		   maxcount : number of start-sites so far
		   workcount: start-sites at positional place in array 
		   locgraph : start-sites ordered and adjacent in array

		   put found new start-pair edges in intermediate array addstart
		   and filter them on the basis of their clan membership and connectivity 
		   to the rest of the nodes in the graph
		 */

		for (i = 0, minlength = M; i < maxcount; i++)
		    /* all graph start-sites */
		  {
		      for (j = 1, fragcount = fragstop = 0; j < dim; j++)
			  if (filter[j])	/* all alignments */
			    {
				for (k = 0, t2 = allstartnum[j]; k < t2; k++)
				  {
				      if (allstart[j][0][k] == locgraph[i]
					  || allstart[j][1][k] == locgraph[i])
					{
					    /* pair is in graph so this alignment in clan for current graph */
					    graphclan[j]++;

					    /* update fragment length for this start-site */
					    fragcount++;
					    if (allstart[j][0][k] == locgraph[i])	/* first in startpair */
						fragstop += scr[j][2];
					    else	/* second in pair */
						fragstop += scr[j][3];
					}
				  }

				if ((t = ali[locgraph[i]][j]) && t >= -M)
				  {
				      for (k = 0, t2 = allstartnum[j];
					   k < t2
					   && allstart[j][0][k] != locgraph[i]
					   && allstart[j][1][k] !=
					   locgraph[i]; k++);
				      if (k >= t2)	/* this pair not yet in alignment[j] startpairs */
					{
					    if (t < 0)
						t = -(t + 1);
					    else
						t--;

					    if ((t3 = ali[t][j]) && t3 >= -M)	/* aligned aa is no gap so new 
										   start-site pair (t, t3) */
					      {
						  if (t3 < 0)
						      t3 = -(t3 + 1);
						  else
						      t3--;

						  if (t3 == locgraph[i])	/* extra ali validity test */
						    {
							if (t < t3)	/* t in upper fragment of ali */
							  {
							      if ((t4 =
								   t -
								   scr[j][2])
								  < virtualvar
								  && t4 >= 0)
								  /* close after ali start */
								  CopyTempStart
								      (addstart,
								       &addstartnum,
								       allstart,
								       allstartnum,
								       -j, t,
								       t3);
							      else
								  CopyTempStart
								      (addstart,
								       &addstartnum,
								       allstart,
								       allstartnum,
								       j, t,
								       t3);
							  }
							else
							  {
							      if ((t4 =
								   t -
								   scr[j][3])
								  < virtualvar
								  && t4 >= 0)
								  /* close after ali start */
								  CopyTempStart
								      (addstart,
								       &addstartnum,
								       allstart,
								       allstartnum,
								       -j, t,
								       t3);
							      else
								  CopyTempStart
								      (addstart,
								       &addstartnum,
								       allstart,
								       allstartnum,
								       j, t,
								       t3);
							  }
						    }
						  else
						    {
							fprintf (stderr,
								 "stacked alignment consistency error at [%ld][%ld].",
								 locgraph[i] +
								 1, j);
							fprintf (stdout, "%c",
								 '\n');
						    }

						  /* now check if new pair includes new start site for 
						     further iteration */
						  for (t2 = 0;
						       t2 < maxcount
						       && locgraph[t2] != t;
						       t2++);
						  if (t2 >= maxcount)	/* t not found so new start-site */
						    {
#ifdef SHOWCLUSTER
							printf
							    ("found new startsite at ali[%ld][%ld]: %ld  (%ld, %ld)\n",
							     locgraph[i] + 1,
							     j, t + 1, t + 1,
							     locgraph[i] + 1);
#endif
						    }
					      }
					}
				  }
			    }

		      fragstop = (fragcount > 0) ? fragstop / fragcount : 0;
		      if (i < maxcount - 1 && fragstop >= locgraph[i + 1])
			  fragstop = locgraph[i + 1] - 1;
		      fragstop = fragstop - locgraph[i] + 1;

		      if (minlength > fragstop)
			  minlength = fragstop;
		  }

#ifdef SHOWCLUSTER
		printf ("===== minlength of graph %ld = %ld\n", graphnum + 1,
			minlength);
#endif


		/* reject new-start sites when they have not enough edges to graph, 
		   when they are too close to end of repeat, or when their edge is not in
		   the graph-clan */

		newstart = 0;	/* number of new start-sites */
		newclan = 0;	/* number of new alignments in clan for this graph */

		for (i = 0; i < addstartnum; i++)	/* all new pairs */
		    if (addstart[0][i])
		      {		/* first filter temp list edges on remaining alignment length */
			  LengthFilter (addstart, i, scr, minfraglen);
		      }

		for (i = 0; i < addstartnum; i++)	/* all new pairs */
		    if (addstart[0][i])
		      {
			  for (j = 0;
			       j < maxcount && addstart[1][i] != locgraph[j];
			       j++);
			  /* all old starts */
			  for (k = 0;
			       k < maxcount && addstart[2][i] != locgraph[k];
			       k++);
			  /* all old starts */
			  if (j < maxcount && k < maxcount)
			      /* both start-sites found so new edge between existing nodes */
			    {
				if (addstart[0][i] < 0)
				    addstart[0][i] = -addstart[0][i];

#ifdef SHOWCLUSTER
				printf
				    ("new edge from alignment %ld between existing nodes [%ld, %ld]\n",
				     addstart[0][i], addstart[1][i] + 1,
				     addstart[2][i] + 1);
#endif
				if (!graphclan[addstart[0][i]])	/* alignment not yet in graphclan */
				    newclan++;
			    }
			  else
			    {
				if (j >= maxcount && k < maxcount)
				    /* not found so addstart[1][i] is new start-site */
				    j = 1;	/* addstart[1][i] = new, locgraph[k] is old */
				else if (k >= maxcount && j < maxcount)
				    /* not found so addstart[2][i] is new start-site */
				    k = j, j = 2;	/* addstart[2][i] = new, locgraph[k] is old */
				else
				  {
				      fprintf (stderr,
					       "alignment %ld pair [%ld, %ld] without node in present graph %ld encountered",
					       addstart[0][i],
					       addstart[1][i] + 1,
					       addstart[2][i] + 1);
				      fprintf (stdout, "%c", '\n');
				  }

				/* count how many connection new start-site has with graph */
				for (t2 = 0, t3 = t4 = 0; t2 < addstartnum;
				     t2++)
				    /* number of start-sites in t3, number of clan start-sites in t4 */
				  {
				      if (addstart[0][t2]
					  && (addstart[1][t2] ==
					      addstart[j][i]
					      || addstart[2][t2] ==
					      addstart[j][i]))
					{
					    t3++;

					    if (addstart[0][t2] < 0)
						t4++;	/* virtual new start-site that is always accepted */
					    else if (graphclan
						     [addstart[0][t2]])
						/* current alignment is in graphclan */
						t4++;
					}
				  }

				/* apply proper graph-clan filter */
				if (clanfilter[0] == 'O')	/* off */
				    t4 = 1;	/* no filtering of start-sites based on graph-clans */
				else if (t4 <= 0 && clanfilter[0] == 'I')	/* inclusive */
				  {
				      for (t = 0; t < dim; t++)	/* stops when start-site is contained in any clan alignment */
					  if (graphclan[t]
					      &&
					      ((scr[t][0] <=
						addstart[j][i] + virtualvar
						&& scr[t][1] >=
						addstart[j][i])
					       || (scr[t][2] <=
						   addstart[j][i] + virtualvar
						   && scr[t][3] >=
						   addstart[j][i])))
					      break;

				      if (t >= dim)	/* start-site is not inclusive so graph-clan test ok */
					{
					    t4 = 1;
#ifdef SHOWCLUSTER
					    printf
						("inferred start-site %ld outside clan alignments so inclusive clan test ok\n",
						 addstart[j][i] + 1);

#endif
					}
#ifdef SHOWCLUSTER
				      else
					  printf
					      ("inferred start-site %ld outside clan alignments so inclusive clan test ok\n",
					       addstart[j][i] + 1);
#endif
				  }

				/* now test for clan and required minimum edge connection */
				if ((addstart[0][i] > 0 || loopcount > 0)
				    && (t4 <= 0
					|| t3 < _Min (maxcount / 2,
						      consconn)))
				    /* no such start-site in clan or 
				       not enough edges so remove from
				       second iteration onward */
				  {
				      for (t2 = i; t2 < addstartnum; t2++)
					  /* remove all pairs containing this site */
					  if (addstart[0][t2]
					      && (addstart[1][t2] ==
						  addstart[j][i]
						  || addstart[2][t2] ==
						  addstart[j][i]))
					    {
#ifdef SHOWCLUSTER
						if (t4 > 0)
						  {
						      printf
							  ("removed infered start-site %ld since connectivity %ld < %ld\n",
							   addstart[j][i] + 1,
							   t3,
							   _Min (maxcount / 2,
								 consconn));
						  }
						else
						  {
						      printf
							  ("removed infered start-site %ld since alignment %ld not in graphclan\n",
							   addstart[j][i] + 1,
							   addstart[0][t2]);
						  }
#endif

						addstart[0][t2] = 0;
					    }
				  }

				/* set scriptnum of virtual startsite now back to positive number */
				if (addstart[0][i] < 0)
				    addstart[0][i] = -addstart[0][i];

				if (addstart[0][i] > 0)
				  {
				      newstart++;
				      if (!graphclan[addstart[0][i]])
					  newclan++;
				  }
			    }
		      }

		/* copy kept startpairs to allstart array */
		CopyStartPair (addstart, addstartnum, allstart, allstartnum,
			       graphnum);

		if (newstart <= 0 && newclan <= 0)
		    fprintf (stdout,
			     "no new start-sites or new alignments found for graph %ld; stop iteration for this graph.\n",
			     graphnum + 1);
		else if (newstart > 0 && newclan <= 0)
		    fprintf (stdout,
			     "new start-sites found for graph %ld: %ld\n",
			     graphnum + 1, newstart);
		else if (newstart <= 0 && newclan > 0)
		    fprintf (stdout,
			     "new alignments found for graph %ld: %ld\n",
			     graphnum + 1, newclan);
		else
		    fprintf (stdout,
			     "graph %ld: new start-sites %ld, new alignments: %ld\n",
			     graphnum + 1, newstart, newclan);

		totnewclan += newclan;
		totnewstart += newstart;
	    }			/* end graphs loop */

	  if (totnewstart <= 0 && totnewclan <= 0)
	      fprintf (stdout, "no new start-sites found; stop iteration.\n");
	  else
	      fprintf (stdout,
		       "iterate for new start-sites: %ld and new alignments: %ld\n",
		       totnewstart, totnewclan);
      }				/* end totnewstart loop */

    maxcount = startgraph[0][0];	/* the largest startgraph */


    /* print the 6 filter parameters */
    fprintf (ofp, "\nGraph-cluster parameters:\n");
    fprintf (ofp, "  minimum required number of edges per graphnode : %ld\n",
	     consconn);
    fprintf (ofp, "  minimum required fragment length : %ld\n", minfraglen);
    fprintf (ofp, "  selection of length of fragments : %s\n", detfraglen);
    fprintf (ofp, "  virtual start-sites allowed shift: %ld\n", virtualvar);
    fprintf (ofp, "  selection of nodes to infer virtual start-sites: %s\n",
	     virtualnodes);
    fprintf (ofp,
	     "  type of filtering of new start-sites based on graph-clans: %s\n",
	     clanfilter);

  /************* initial startgraphs printed ***********/
    initadjmat = (long ***) ckalloc (initgraphamount * sizeof (long **),
				     "init adjmat array");
    initallstartnum = (long *) ckalloc (dim * sizeof (long),
					"dummy allstartnum array");
    for (t = 1; t < dim; t++)
	initallstartnum[t] = 1;	/* only first start-pair
				   of each ali taken */
    MakeAdjMats (allstart, initallstartnum, initstartgraph, initgraphamount,
		 initadjmat, dim, filter);
    /* now make dummy subgraphlist for printing the complete non-decomposed
       graphs */
    lsubgraph = (S_Subgraph *) ckalloc (initgraphamount * sizeof (S_Subgraph),
					"array of subgraphs");
    for (i = 0; i < initgraphamount; i++)
	lsubgraph[i].graphid = i,
	    lsubgraph[i].subgraphid = 0,
	    lsubgraph[i].subgraphnum = initstartgraph[i][0];
    /* make dummy subgraph matrix */
    alloc_2d (subgraph, long, initgraphamount, (maxcount + 1),
	      "subgraphs  array");

    strcpy (graphtitle, "Initial start-graph");
    PrintGraph (ofp, initgraphamount, initstartgraph, subgraph, lsubgraph,
		initadjmat, wght, lenwght, dim, scr, NULL, graphtitle,
		detfraglen);

    ckfree (lsubgraph);
    free_2d (subgraph);
    for (t = 0; t < initgraphamount; t++)
	ckfree/*_2d*/ (initadjmat[t][0]);
    ckfree (initadjmat);
    free_2d (initstartgraph);
  /************* end of initial startgraphs printed ***********/

    adjmat =
	(long ***) ckalloc (graphamount * sizeof (long **), "adjmat array");

    MakeAdjMats (allstart, allstartnum, startgraph, graphamount, adjmat, dim,
		 filter);

    /* make subgraph matrix */
    subgraph = (long **) ckalloc (graphamount * sizeof (long *),
				  "subgraphs array");
    subgraph[0] =
	(long *) ckcalloc (graphamount * (maxcount + 1), sizeof (long),
			   "subgraphs data array");
    for (t = 1; t < graphamount; t++)
	subgraph[t] = subgraph[t - 1] + maxcount + 1;

    /* now make dummy subgraphlist for printing the complete non-decomposed
       graphs */
    lsubgraph = (S_Subgraph *) ckalloc (graphamount * sizeof (S_Subgraph),
					"array of subgraphs");
    for (i = 0; i < graphamount; i++)
	lsubgraph[i].graphid = i,
	    lsubgraph[i].subgraphid = 0,
	    lsubgraph[i].subgraphnum = startgraph[i][0];
    strcpy (graphtitle, "Non-decomposed start-graph");
    PrintGraph (ofp, graphamount, startgraph, subgraph, lsubgraph, adjmat,
		wght, lenwght, dim, scr, NULL, graphtitle, detfraglen);
    ckfree (lsubgraph);


    Decompose (startgraph, graphamount, adjmat, subgraph);

    for (i = 0, subgraphnum = 0; i < graphamount; i++)
      {
	  subgraphnum += subgraph[i][0];
	  /*
	     printf("Start-graph: %ld,  %ld subgraphs\n", i+1, subgraph[i][0]);
	     for (j = 0; j < startgraph[i][0]; j++)
	     printf("%4ld.  %4ld  %4ld\n", j+1, startgraph[i][j+1]+1, 
	     subgraph[i][j+1]);
	     printf("\n%s%s\n",
	     "-------------------------------------------------------------",
	     "-------------------------------------------------------------");
	   */
      }

    lsubgraph = (S_Subgraph *) ckalloc (subgraphnum * sizeof (S_Subgraph),
					"array of subgraphs");

    ListSubgraphs (subgraph, lsubgraph, graphamount, subgraphnum, startgraph);

    frags = (long **) ckalloc (subgraphnum * sizeof (long *),
			       "subgraphs seq fragments array");
    frags[0] =
	(long *) ckcalloc (subgraphnum * (k = lsubgraph[0].subgraphnum * 2),
			   sizeof (long),
			   "subgraphs seq fragments data array");

    for (t = 1; t < subgraphnum; t++)
	frags[t] = frags[t - 1] + k;

    strcpy (graphtitle, "Start-subgraph");
    PrintGraph (ofp, subgraphnum, startgraph, subgraph, lsubgraph, adjmat,
		wght, lenwght, dim, scr, frags, graphtitle, detfraglen);

    PrintFrags (ofp, A, frags, lsubgraph, subgraphnum);


    ckfree (wght);
    ckfree (lenwght);
    ckfree (workcount);
    ckfree (workwght);
    ckfree (start);
    ckfree (scr[0]);
    ckfree (scr);
    free_3d (allstart);
    ckfree (oldallstartnum);
    ckfree (allstartnum);
    ckfree (startgraph[0]);
    ckfree (startgraph);
    /* ckfree(locgraph); */
    ckfree (addstart[0]);
    ckfree (addstart);
    ckfree (graphclan);
    ckfree (subgraph[0]);
    ckfree (subgraph);
    for (t = 0; t < graphamount; t++)
	/*ckfree (adjmat[t][0][0]),*/ ckfree (adjmat[t][0]);
    ckfree (adjmat);
    ckfree (lsubgraph);
    ckfree (frags[0]);
    ckfree (frags);
    ckfree (initallstartnum);
}


/*----------------------------------------------------------------*/

void LengthFilter (long **addstart, long i, long **scr, long minfraglen)
{
    long t2, t3, t4, alinum, j;

    /* delete start-sites just before end of top alignment if remaining stretch
       is shorter than minfraglen   --- code change 8-2-1995 ---
     */

    t2 = (addstart[0][i] >= 0) ? addstart[0][i] : -addstart[0][i];
    /* virtual startpoint has neg. alignment number */

    alinum = addstart[0][i];

    t3 = scr[t2][2] - addstart[1][i];
    t4 = scr[t2][3] - addstart[2][i];
    if (t4 < t3)
	t3 = t4, j = 2;
    else
	j = 1;
    /* minimal distance between the two 
       start-points and physical end-points 
       scr[..][2] or scr[..][3] */
    /*
       if (t3 < maxfraglen && (t3 <= minlength/2 || t3 < minfraglen)) 
       --- code change 8-2-1995 ---
     */

    if (t3 < minfraglen)
	/* too close before end so delete */
      {
	  addstart[0][i] = 0;
#ifdef SHOWCLUSTER
	  printf
	      ("removed start-site %ld in alignment %ld (%ld, %ld)  since %ld before end\n",
	       addstart[j][i] + 1, t2, addstart[1][i] + 1, addstart[2][i] + 1,
	       t3);
#endif
      }
}

/*----------------------------------------------------------------*/

void
CopyCloseStarts (long **startgraph, long graphnum, long ***allstart,
		 long *allstartnum, long **addstart, long *addstartnum_ptr,
		 char *filter, long scriptsdim, long testfac, long **scr,
		 long virtualvar, char *virtualnodes)
{
/* CopyCloseStart finds virtual startsites in scripts that have a start-site
 * starting less/equal virtualvar residues after one or two startsites already
 * present in current startgraph; if only close after one start-site then a
 * new virtual start-site is found, if both starts come close after two
 * start-sites in the current graph then these are also taken as they allow
 * current script to be included in the clan for this graph (and thus more
 * start-sites can be infered from this script in subsequent iterations)  
 */

    long i, j, t, t2, t3, t4, t5, diff1, diff2, p1, p2, vn;

    t4 = startgraph[graphnum][0];

    for (i = 1; i < scriptsdim; i++)	/* pairs array loop */
	if (filter[i])
	  {
	      t = allstartnum[i];

	      for (j = 0, vn = (virtualnodes[0] == 'A') ? allstartnum[i] : 1; j < vn; j++)	/* all pairs in current script */
		{
		    for (t2 = 1, p1 = -10000; t2 <= t4; t2++)	/* all graphnodes */
		      {
			  if ((diff1 =
			       allstart[i][0][j] -
			       startgraph[graphnum][t2]) <= virtualvar
			      && diff1 > 0)
			      /* startside close after current graph start-site, so add
			         new start-site, by taking same difference from the
			         other start-site (a bit dangerous, maybe we should do
			         Needleman-Wunsch again) */
			    {
				p1 = allstart[i][0][j] - diff1;
				break;
			    }
		      }

		    for (t3 = 1, p2 = -10000; t3 <= t4; t3++)
			if ((diff2 =
			     allstart[i][1][j] - startgraph[graphnum][t3]) <=
			    virtualvar && diff2 > 0)
			  {
			      p2 = allstart[i][1][j] - diff2;
			      break;
			  }

		    if (p1 >= 0 && p2 > scr[i][2])	/* p2 should not overlap with first
							   segment of ali */
		      {
			  if (diff1 == diff2)
			    {
				/* check if pair is already there for this script */
				for (t5 = 0; t5 < t && allstart[i][0][t5] != p1 && allstart[i][1][t5] != p2; t5++)	/* all pairs in current alignment */
				    ;

				if (t5 >= t)	/* pair not found */
				  {
#ifdef SHOWCLUSTER
				      printf
					  ("found virtual existing startpair [%ld, %ld] in script %ld\n",
					   p1 + 1, p2 + 1, i);
#endif
				      CopyTempStart (addstart,
						     addstartnum_ptr,
						     allstart, allstartnum, i,
						     p1, p2);
				  }
			    }
		      }
		    else if ((p1 >= 0 && p2 < 0)
			     || (p1 < 0 && p2 > scr[i][2]))
			/* new start p2 or p1 found */
		      {
			  if (p1 >= 0)
			      p2 = allstart[i][1][j] - diff1;
			  else
			      p1 = allstart[i][0][j] - diff2;

			  if (p1 >= 0 && p2 > scr[i][2])	/* both start in sequence */
			    {
				/* check if pair is already there for this script */
				for (t5 = 0; t5 < t && ((allstart[i][0][t5] != p1 && allstart[i][1][t5] != p2) || (allstart[i][0][t5] != p2 && allstart[i][1][t5] != p1)); t5++)	/* all pairs in current script */
				    ;

				if (t5 >= t)	/* pair not found */
				  {
#ifdef SHOWCLUSTER
				      printf
					  ("found virtual startpair [%ld, %ld] in script %ld\n",
					   p1 + 1, p2 + 1, i);
#endif
				      /* set scriptno to testfac*i to recognize this situation */
				      CopyTempStart (addstart,
						     addstartnum_ptr,
						     allstart, allstartnum,
						     testfac * i, p1, p2);
				  }
			    }
		      }
		}
	  }
}

/*----------------------------------------------------------------*/

void
CopyStartPair (long **addstart, long addstartnum, long ***allstart,
	       long *allstartnum, long graphnum)
{
    /* copy kept startpairs to allstart array */

    long i, t2, t3, t4, s1, s2;

    for (i = 0; i < addstartnum; i++)	/* all new pairs */
	if ((t2 = addstart[0][i]))
	  {
	      if ((t3 = allstartnum[t2]) < ALLSTARTLIM)
		{
		    if (addstart[1][i] < addstart[2][i])
			s1 = addstart[1][i], s2 = addstart[2][i];
		    else
			s2 = addstart[1][i], s1 = addstart[2][i];

		    for (t4 = 0; t4 < t3
			 && allstart[t2][0][t4] != s1
			 && allstart[t2][1][t4] != s2; t4++)
			;
		    if (t4 >= t3)	/* pair not existing already */
		      {
			  allstart[t2][0][t3] = s1;
			  allstart[t2][1][t3] = s2;
			  allstartnum[t2]++;

#ifdef SHOWCLUSTER
			  printf
			      ("copied start-site[%ld](%ld) (%ld, %ld) to allstart array\n",
			       t2, allstartnum[t2], allstart[t2][0][t3] + 1,
			       allstart[t2][1][t3] + 1);
#endif
		      }
#ifdef SHOWCLUSTER
		    else
		      {
			  printf
			      ("CopyStartPair: start-site [%ld]->(%ld, %ld) not copied: exists already\n",
			       t2, allstartnum[t2], allstart[t2][0][t3] + 1,
			       allstart[t2][1][t3] + 1);
		      }
#endif
		}
	      else
		{
		    printf
			("upper limit %d reached for start-sites in repeat %d.\n",
			 ALLSTARTLIM, t2);
		    exit (0);
		}
	  }
}

/*----------------------------------------------------------------*/

void
CopyTempStart (long **addstart, long *addstartnum_ptr,
	       long ***allstart, long *allstartnum,
	       long scriptnum, long p1, long p2)
{
    long t, t2, scriptnum2;

    if (*addstartnum_ptr < TEMPSTARTLIM)
      {
	  scriptnum2 = (scriptnum > 0) ? scriptnum : -scriptnum;

	  if (p1 > p2)
	      t = p1, p1 = p2, p2 = t;	/* reverse (p1, p2) */

	  for (t = 0; t < *addstartnum_ptr; t++)	/* check if this pair is already in list */
	      if ((_Abs (addstart[0][t])) == scriptnum2
		  && addstart[1][t] == p1 && addstart[2][t] == p2)
		  break;

	  if (t >= *addstartnum_ptr)	/* not found */
	    {
		for (t = 0, t2 = allstartnum[scriptnum2];
		     t < t2
		     && allstart[scriptnum2][0][t] != p1
		     && allstart[scriptnum2][1][t] != p2; t++)
		    ;

		if (t >= t2)	/* pair not existing already */
		  {
		      addstart[0][*addstartnum_ptr] = scriptnum,
			  addstart[1][*addstartnum_ptr] = p1,
			  addstart[2][*addstartnum_ptr] = p2;

#ifdef SHOWCLUSTER
		      printf
			  ("copied start-site [%ld]->(%ld, %ld) to temporary pairs list\n",
			   scriptnum, p1 + 1, p2 + 1);
#endif

		      (*addstartnum_ptr)++;
		  }
		else
		  {
#ifdef SHOWCLUSTER
		      printf
			  ("start-site [%ld]->(%ld, %ld) not copied: exists already in allstart array\n",
			   scriptnum, p1 + 1, p2 + 1);
#endif
		  }
	    }
#ifdef SHOWCLUSTER
	  else
	      printf
		  ("start-site [%ld]->(%ld, %ld) exists already in temporary pairs list\n",
		   scriptnum, p1 + 1, p2 + 1);
#endif
      }
    else
      {
	  printf ("upper limit %d reached for new starts.\n", TEMPSTARTLIM);
	  exit (0);
      }
}

/*----------------------------------------------------------------*/

void MakeAdjMats (long ***allstart, long *allstartnum, long **startgraph, long graphnum, long ***adjmat, long dim, char *filter)
{
    long i, j, k, t, t2, t3, t4, *locgraph, **locadjmat = NULL;

    for (i = 1; i < dim; i++)
      {
	  fprintf (stdout, "\nalignment %ld: start-pairs: %ld", i, k =
		   allstartnum[i]);
	  for (j = 0; j < k; j++)
	    {
		fprintf (stdout, "%s[%3ld,%3ld]", (!(j % 10)) ? "\n  " : "  ",
			 allstart[i][0][j] + 1, allstart[i][1][j] + 1);
	    }
	  fprintf (stdout, "%c", '\n');
      }

    for (i = 0; i < graphnum; i++)
      {
	  locadjmat =
	      (long **) ckalloc ((t = startgraph[i][0]) * sizeof (long *),
				 "graphclan ptrs array");
	  locadjmat[0] =
	      (long *) ckcalloc (t * t, sizeof (long),
				 "graphclan data array");
	  adjmat[i] = locadjmat;

	  for (j = 1; j < t; j++)
	      locadjmat[j] = locadjmat[j - 1] + t;

	  locgraph = &(startgraph[i][1]);

	  for (t2 = 1; t2 < dim; t2++)	/* loop over alignments */
	      if (filter[t2])
		{
		    for (t3 = 0, t4 = allstartnum[t2]; t3 < t4; t3++)
			for (j = 0; j < t; j++)
			  {
			      if (locgraph[j] == allstart[t2][0][t3]
				  || locgraph[j] == allstart[t2][1][t3])
				{
				    /* update adjacency matrix */
				    if (locgraph[j] == allstart[t2][0][t3])
				      {
					  for (k = 0;
					       k < t
					       && locgraph[k] !=
					       allstart[t2][1][t3]; k++)
					      ;
					  if (k < t)	/* found */
					      locadjmat[j][k] =
						  locadjmat[k][j] = t2;
					  else
					    {
						fprintf (stderr,
							 "pair [%ld, %ld] in alignment %ld: %ld not in graph %ld",
							 allstart[t2][0][t3],
							 allstart[t2][1][t3],
							 t2,
							 allstart[t2][1][t3],
							 i + 1);
						fprintf (stdout, "%c", '\n');
					    }
				      }
				    else
				      {
					  for (k = 0;
					       k < t
					       && locgraph[k] !=
					       allstart[t2][0][t3]; k++)
					      ;
					  if (k < t)	/* found */
					      locadjmat[j][k] =
						  locadjmat[k][j] = t2;
					  else
					    {
						fprintf (stderr,
							 "pair [%ld, %ld] in alignment %ld: %ld not in graph %ld",
							 allstart[t2][0][t3],
							 allstart[t2][1][t3],
							 t2,
							 allstart[t2][0][t3],
							 i + 1);
						fprintf (stdout, "%c", '\n');
					    }
				      }
				}
			  }
		}
      }
}


/*---------------------------------------------------------------------*/

void
PrintGraph (FILE * ofp, long graphsize, long **startgraph, long **subgraph,
	    S_Subgraph * lsubgraph, long ***adjmat, long *aliwght,
	    long *alilenwght, long alinum, long **alistart, long **frags,
	    char *graphtitle, char *detfraglen)
{
    long i, j, k, *locgraph, totedges, totwghtedges, totlenwghtedges,
	graphloop, graphnum, subgraphnum, *locsubgraph, **locadjmat,
	graphid, aliid, *helpali, *edges, *wghtedges, *lenwghtedges,
	subgraphid;

    helpali = (long *) ckcalloc (alinum, sizeof (long), "help ali array");

    for (graphloop = 0; graphloop < graphsize; graphloop++)
      {
	  graphnum = startgraph[(graphid = lsubgraph[graphloop].graphid)][0];
	  assert(graphnum > 0);

	  memset(helpali, 0, alinum * sizeof (long));
	  edges = ckcalloc(graphnum, sizeof (long), "help edges array");
	  wghtedges = ckcalloc(graphnum, sizeof (long), "help wghtedges array");
	  lenwghtedges = ckcalloc(graphnum, sizeof (long), "help lenwghtedges array");

	  subgraphnum = lsubgraph[graphloop].subgraphnum;
	  subgraphid = lsubgraph[graphloop].subgraphid;
	  locadjmat = adjmat[graphid];
	  locgraph = &startgraph[graphid][1];
	  locsubgraph = &subgraph[graphid][1];

	  for (i = 0, totedges = totwghtedges = totlenwghtedges = 0;
	       i < graphnum; i++)
	      if (locsubgraph[i] == subgraphid)
		{
		    for (j = 0; j < graphnum; j++)
			if (locsubgraph[j] == subgraphid
			    && (aliid = locadjmat[i][j]) > 0)
			    edges[i]++, wghtedges[i] +=
				aliwght[aliid], lenwghtedges[i] +=
				alilenwght[aliid], helpali[aliid]++;

		    totedges += edges[i];
		    totwghtedges += wghtedges[i];
		    totlenwghtedges += lenwghtedges[i];
		}

	  if (subgraphnum > 1)
	      fprintf (ofp,
		       "\n%s %ld-%ld: %ld %s\n%s %.2f, %s %.2f, %s %.2f\n%s %.2f, %s %.2f, %s %.2f\n",
		       graphtitle,
		       graphid + 1, subgraphid,
		       subgraphnum,
		       "members",
		       "mean connectivity:",
		       1.0 * totedges / subgraphnum,
		       "mean weighted connectivity:",
		       RVMULT * totwghtedges / subgraphnum,
		       "mean score-weighted connectivity:",
		       RVMULT * totlenwghtedges / subgraphnum,
		       "mean normalized connectivity:",
		       1.0 * totedges / (subgraphnum * (subgraphnum - 1)),
		       "mean weighted normalized connectivity:",
		       RVMULT * totwghtedges / (subgraphnum *
						(subgraphnum - 1)),
		       "mean score-weighted normalized connectivity:",
		       RVMULT * totlenwghtedges / (subgraphnum *
						   (subgraphnum - 1)));
	  else
	      fprintf (ofp,
		       "\n%s %ld-%ld: %ld %s\n%s %.2f, %s %.2f, %s %.2f\n",
		       graphtitle,
		       graphid + 1, subgraphid,
		       subgraphnum,
		       "members",
		       "mean connectivity:",
		       (subgraphnum > 0) ? 1.0 * totedges / subgraphnum : 0,
		       "mean weighted connectivity:",
		       (subgraphnum >
			0) ? RVMULT * totwghtedges / subgraphnum : 0,
		       "mean score-weighted connectivity:",
		       (subgraphnum >
			0) ? RVMULT * totlenwghtedges / subgraphnum : 0);

	  fprintf (ofp, "%10s%7s%12s  ", "Start-site", "conn.", "wght conn.");

	  for (i = 1; i < alinum; i++)
	      if (helpali[i])
		  fprintf (ofp, " %3s%2ld", "Ali", i);
	  fprintf (ofp, "%c", '\n');

	  for (i = 0; i < graphnum; i++)
	      if (locsubgraph[i] == subgraphid)
		{
		    fprintf (ofp, "%10ld%7ld%12.2f  ",
			     locgraph[i] + 1, edges[i],
			     RVMULT * wghtedges[i]);

		    for (j = 1; j < alinum; j++)
			if (helpali[j])
			  {
			      for (k = 0; k < graphnum
				   && (locsubgraph[k] != subgraphid
				       || locadjmat[i][k] != j); k++)
				  ;
			      fprintf (ofp, " %5s",
				       ((k < graphnum) ? "X" : "-"));
			  }
		    fprintf (ofp, "%c", '\n');
		}

	  fprintf (ofp, "\n%s %ld-%ld: repeat endpoints:\n", graphtitle,
		   graphid + 1, subgraphid);
	  fprintf (ofp, "%10s  ", "Start-site");

	  for (i = 1; i < alinum; i++)
	      if (helpali[i])
		  fprintf (ofp, " %3s%2ld", "Ali", i);
	  fprintf (ofp, "%c", '\n');

	  for (i = 0; i < graphnum; i++)
	      if (locsubgraph[i] == subgraphid)
		{
		    fprintf (ofp, "%10ld  ", locgraph[i] + 1);

		    for (j = 1; j < alinum; j++)
		      {
			  if (helpali[j])
			    {
				for (k = 0; k < graphnum
				     && (locsubgraph[k] != subgraphid
					 || locadjmat[i][k] != j); k++)
				    ;
				if (k < graphnum)	/* ali is there */
				  {
				      if (locgraph[i] < alistart[j][2])	/* in first fragment of ali */
					  fprintf (ofp, " %5ld",
						   alistart[j][2] + 1);
				      else	/* in second fragment of this ali */
					  fprintf (ofp, " %5ld",
						   alistart[j][3] + 1);
				  }
				else
				    fprintf (ofp, " %5s", "-");
			    }
			  else if (graphnum == 1)
			      fprintf (ofp, " %5s", "-");
		      }
		    fprintf (ofp, "%c", '\n');
		}



	  /* determine  fragment endpoints according to fraglenind */
	  DetFragEnds (frags, detfraglen, graphloop, graphnum, locgraph,
		       locsubgraph, subgraphid, alinum, alistart, helpali,
		       locadjmat);

	  fprintf (ofp,
		   "\n\n%s %ld-%ld adjacency matrix with alignment indicators:\n",
		   graphtitle, graphid + 1, subgraphid);
	  fprintf (ofp, "%4s ", " ");
	  for (j = 0; j < graphnum; j++)
	      if (locsubgraph[j] == subgraphid)
		  fprintf (ofp, " %3ld", locgraph[j] + 1);
	  fprintf (ofp, "%c", '\n');
	  for (i = 0; i < graphnum; i++)
	      if (locsubgraph[i] == subgraphid)
		{
		    fprintf (ofp, "%-4d ", locgraph[i] + 1);
		    for (j = 0; j < graphnum; j++)
			if (locsubgraph[j] == subgraphid)
			    fprintf (ofp, " %3ld", locadjmat[i][j]);
		    fprintf (ofp, "%c", '\n');
		}

	  fprintf (ofp, "\n%s%s\n",
		   "-------------------------------------------------------------",
		   "-------------------------------------------------------------");


	  ckfree (edges);
	  ckfree (wghtedges);
      }

    ckfree (helpali);
}

/*----------------------------------------------------------------*/

void
DetFragEnds (long **frags, char *detfraglen, long graphloop, long graphnum,
	     long *locgraph, long *locsubgraph, long subgraphid, long alinum,
	     long **alistart, long *helpali, long **locadjmat)
{
    long fraglenind, i, j, k, t, sum_end, n_end, fragid;


    /* set internal integer with frags endpoints determination codes */

    if (!strcmp (detfraglen, "MEANL"))
	fraglenind = 0;		/*default */
    else
      {
	  if (!strcmp (detfraglen, "MAXL"))
	      fraglenind = 1;	/* take longest ali-endp */
	  else
	    {
		fprintf (stderr, "detfraglen: %s\n", detfraglen);
		fatal ("invalid indicator for fragment endpoints");
	    }
      }

    if (frags != NULL && fraglenind == 0)	/* default mean fragment endpoint */
      {
	  for (i = fragid = 0; i < graphnum; i++)
	      if (locsubgraph[i] == subgraphid)
		{
		    for (j = 1, sum_end = n_end = 0; j < alinum; j++)
		      {
			  if (helpali[j])
			    {
				for (k = 0; k < graphnum
				     && (locsubgraph[k] != subgraphid
					 || locadjmat[i][k] != j); k++)
				    ;
				if (k < graphnum)	/* ali is there */
				  {
				      if (locgraph[i] < alistart[j][2])	/* in first fragment of ali */
					{
					    for (t = i + 1; t < graphnum; t++)
						if (locsubgraph[t] ==
						    subgraphid)
						    break;
					    if (t < graphnum)	/* there is a start-site following this one */
					      {
						  if (alistart[j][2] < locgraph[t])	/* ali is shorter than 
											   next start-site */
						      sum_end +=
							  alistart[j][2];
						  else
						      sum_end +=
							  locgraph[t] - 1;
					      }
					    else
						sum_end += alistart[j][2];

					    n_end++;
					}
				      else	/* in second fragment of this ali */
					{
					    for (t = i + 1; t < graphnum; t++)
						if (locsubgraph[t] ==
						    subgraphid)
						    break;
					    if (t < graphnum)	/* there is a start-site following this one */
					      {
						  if (alistart[j][3] < locgraph[t])	/* ali is shorter than 
											   next start-site */
						      sum_end +=
							  alistart[j][3];
						  else
						      sum_end +=
							  locgraph[t] - 1;
					      }
					    else
						sum_end += alistart[j][3];

					    n_end++;
					}
				  }
			    }
		      }

		    frags[graphloop][fragid++] = locgraph[i];
		    frags[graphloop][fragid++] =
			(n_end > 0) ? sum_end / n_end : 0;
		}
      }
    else if (frags != NULL && fraglenind == 1)
	/* take length of longest alignment for fragment */
      {
	  for (i = fragid = 0; i < graphnum; i++)
	      if (locsubgraph[i] == subgraphid)
		{
		    for (j = 1, sum_end = 0; j < alinum; j++)
		      {
			  if (helpali[j])
			    {
				for (k = 0; k < graphnum
				     && (locsubgraph[k] != subgraphid
					 || locadjmat[i][k] != j); k++)
				    ;
				if (k < graphnum)	/* ali is there */
				  {
				      if (locgraph[i] < alistart[j][2])	/* in first fragment of ali */
					{
					    for (t = i + 1; t < graphnum; t++)
						if (locsubgraph[t] ==
						    subgraphid)
						    break;
					    if (t < graphnum)	/* there is a start-site following this one */
					      {
						  if (alistart[j][2] < locgraph[t])	/* ali is shorter than 
											   next start-site */
						    {
							if (sum_end <
							    alistart[j][2])
							    sum_end =
								alistart[j]
								[2];
						    }
						  else
						    {
							if (sum_end <
							    locgraph[t] - 1)
							    sum_end =
								locgraph[t] -
								1;
						    }
					      }
					    else if (sum_end < alistart[j][2])
						sum_end = alistart[j][2];
					}
				      else	/* in second fragment of this ali */
					{
					    for (t = i + 1; t < graphnum; t++)
						if (locsubgraph[t] ==
						    subgraphid)
						    break;
					    if (t < graphnum)	/* there is a start-site following this one */
					      {
						  if (alistart[j][3] < locgraph[t])	/* ali is shorter than 
											   next start-site */
						    {
							if (sum_end <
							    alistart[j][3])
							    sum_end =
								alistart[j]
								[3];
						    }
						  else
						    {
							if (sum_end <
							    locgraph[t] - 1)
							    sum_end =
								locgraph[t] -
								1;
						    }
					      }
					    else if (sum_end < alistart[j][3])
						sum_end = alistart[j][3];
					}
				  }
			    }
		      }

		    frags[graphloop][fragid++] = locgraph[i];
		    frags[graphloop][fragid++] = sum_end;	/* contains max endp */
		}
      }
}

/*----------------------------------------------------------------*/

void PrintFrags (FILE * ofp, char A[], long **frags, S_Subgraph * lsubgraph, long subgraphnum)
{
    long i, j, k, t, lim, start, stop;

    fprintf (ofp, "\nsequence fragments associated to start-graphs:\n\n");

    for (i = 0; i < subgraphnum; i++)
      {
	  if ((lim = lsubgraph[i].subgraphnum) <= 1)
	      continue;

	  fprintf (ofp, "start_graph %ld-%ld: %ld fragments\n",
		   lsubgraph[i].graphid + 1, lsubgraph[i].subgraphid,
		   lim = lsubgraph[i].subgraphnum);

	  lim *= 2;

	  for (j = 0; j < lim; j += 2)
	    {
		fprintf (ofp, ">F1;%ld%ld\n", (start = frags[i][j]) + 1,
			 (stop = frags[i][j + 1]) + 1);
		fprintf (ofp,
			 "repeat fragment containing residues %ld to %ld",
			 start + 1, stop + 1);
		for (k = start, t = 0; k <= stop; k++)
		    fprintf (ofp, "%s%c", (!(t++ % 50)) ? "\n" : "", A[k]);
		fprintf (ofp, "*%c", '\n');
	    }
	  fprintf (ofp, "%s%s\n",
		   "----------------------------------------",
		   "----------------------------------------");
      }
}

/*----------------------------------------------------------------*/

void Decompose (long **startgraph, long graphnum, long ***adjmat, long **subgraph)
{
    long num, i, j, k, count, id, maxcount, maxid, *helpgraph, graphid,
	**locadjmat,
	oldgraphid, save_edge, *locsubgraph, *helpgraph2, subcount, graphloop;

    for (graphloop = 0; graphloop < graphnum; graphloop++)
      {
	  if ((num = startgraph[graphloop][0]) < 3)
	    {
		subgraph[graphloop][0] = 1;	/* number of decomposed subgraphs in this
						   graph */
		for (i = 1; i <= num; i++)
		    subgraph[graphloop][i] = 1;
	    }
	  else
	    {
		locadjmat = adjmat[graphloop];
		locsubgraph = &(subgraph[graphloop][1]);

		helpgraph = (long *) ckalloc (num * sizeof (long),
					      "helpgraph decomp. array");
		helpgraph2 = (long *) ckcalloc (num, sizeof (long),
						"helpgraph2 decomp. array");

		for (i = 0, graphid = 0; i < num; i++)
		    for (j = i + 1; j < num; j++)
			if (locadjmat[i][j])
			  {
			      memset (helpgraph, 0, num * sizeof (long));
			      helpgraph[i] = 1;
			      save_edge = locadjmat[i][j];
			      locadjmat[i][j] = locadjmat[j][i] = 0;

			      testdecomp (i, num, locadjmat, locsubgraph,
					  helpgraph);

			      locadjmat[i][j] = locadjmat[j][i] = save_edge;

			      if (!helpgraph[j])	/* graph is decomposable by edge [i, j] */
				{
				    fprintf (stdout,
					     "graph %ld decomposable by edge [%ld: %ld, %ld: %ld]\n",
					     graphloop + 1, i + 1,
					     startgraph[graphloop][i + 1] + 1,
					     j + 1,
					     startgraph[graphloop][j + 1] +
					     1);
				    for (k = 0, graphid += 2, oldgraphid =
					 locsubgraph[i]; k < num; k++)
					if (locsubgraph[k] == oldgraphid)
					  {
					      if (helpgraph[k])
						  locsubgraph[k] =
						      graphid - 1;
					      else
						  locsubgraph[k] = graphid;
					  }
				}
			  }


		/* convert subgraph identification numbers to id's representing order
		   according to number of members of subgraphs */
		subcount = graphid = 0;
		while (subcount < num)
		  {
		      memset (helpgraph, 0, num * sizeof (long));
		      for (i = 0, maxcount = 0; i < num; i++)
			{
			    if (!helpgraph[i] && !helpgraph2[i])
			      {
				  id = locsubgraph[i];
				  for (j = i, count = 0; j < num; j++)
				    {
					if (locsubgraph[j] == id)
					    count++, helpgraph[j] = id;
				    }
				  if (count > maxcount)
				      maxcount = count, maxid = id;
			      }
			}

		      /* helpgraph2 will contain oredered subgraph id's */
		      for (i = 0, graphid++, subcount += maxcount; i < num;
			   i++)
			  if (locsubgraph[i] == maxid)
			      helpgraph2[i] = graphid;
		  }

		assert(subcount == num);

		subgraph[graphloop][0] = graphid;	/* number of decomposed subgraphs in this
							   graph */
		/* copy id's to subgraph */
		for (i = 0; i < num; i++)
		    locsubgraph[i] = helpgraph2[i];

		ckfree (helpgraph);
		ckfree (helpgraph2);
	    }
      }
}

/*----------------------------------------------------------------*/

int SortGraph (const void *ptr1, const void *ptr2)
{
    const S_Subgraph *p1 = ptr1, *p2 = ptr2;
    int		     diff;

    if ((diff = p2->subgraphnum - p1->subgraphnum) != 0 ||
	(diff = p1->graphid - p2->graphid) != 0 ||
	(diff = p1->subgraphid - p2->subgraphid) != 0)
	return diff;

    assert(0);/*fatal ("Internal error in sorting subgraphs\n");*/
}

/*----------------------------------------------------------------*/

void ListSubgraphs (long **subgraph, S_Subgraph * lsubgraph, long graphsize, long subgraphsize, long **startgraph)
{
/*
typedef struct graph {long graphid, subgraphid, subgraphnum;} 
     S_Subgraph;
*/
    long i, j, k, lim1, lim2, count, *locsubgraph, graphnum;

    for (i = 0, graphnum = 0; i < graphsize; i++)
      {
	  for (j = 1, lim1 = subgraph[i][0], locsubgraph = &(subgraph[i][1]);
	       j <= lim1; j++, graphnum++)
	    {
		lsubgraph[graphnum].graphid = i;
		lsubgraph[graphnum].subgraphid = j;

		for (k = 0, lim2 = startgraph[i][0], count = 0; k < lim2; k++)
		    if (locsubgraph[k] == j)
			count++;
		lsubgraph[graphnum].subgraphnum = count;
	    }
      }

    assert(graphnum == subgraphsize);
    qsort (lsubgraph, subgraphsize, sizeof (S_Subgraph), SortGraph);

     /**/ printf ("\nordered subgraphs list:\n");
    for (i = 0; i < subgraphsize; i++)
	printf ("%4ld. %3ld-%3ld:  %3ld nodes\n", i + 1,
		lsubgraph[i].graphid + 1, lsubgraph[i].subgraphid,
		lsubgraph[i].subgraphnum);
 /**/}

/*----------------------------------------------------------------*/

void testdecomp (long node1, long dim, long **adjmat, long *subgraph, long *helpgraph)
{
    long i, subgraphid;
/*
  static long rec = 0;

  for (i = 0, rec++; i < rec; i++) printf("X");
  printf(" %ld\n", node1);
*/

    for (i = 0, subgraphid = subgraph[node1]; i < dim; i++)
	if (subgraphid == subgraph[i] && adjmat[node1][i] && node1 != i)
	  {
	      if (!helpgraph[i])
		{
		    helpgraph[i]++;
		    testdecomp (i, dim, adjmat, subgraph, helpgraph);
		}
	      else
		  helpgraph[i]++;
	  }

/*
  rec--;
*/
}

/*----------------------------------------------------------------*/




/******************* lib.c - library of C procedures. *************************/

const char *prefix(const char *fn, char *s)
{
    long i, j, k;
    char *s2;

    for (i = strlen(fn) - 1;
	 i >= 0 && fn[i] != '/' && fn[i] != ']' && fn[i] != ':'
	 && fn[i] != '\\'; i--);
    /* this should cover most operating systems like UNIX, VMS or DOS */
    i++;

    for (j = i; fn[j] && fn[j] != '.'; j++);

    for (k = i, s2 = s; k < j; *(s2++) = fn[k++]);

    *s2 = '\0';

    return s;
}


/* fatal - print message and die */
void fatal(const char *msg)
{
    fprintf (stderr, "%s\n", msg);
    exit (1);
}

/* ckopen - open file; check for success */
FILE *ckopen(const char *name, const char *mode)
{
    FILE *fp;

    if ((fp = fopen(name, mode)) == NULL) {
	fprintf(stderr, "Cannot open %s\n", name);
	exit(1);
    }

    return fp;
}


/* ckalloc - allocate space; align on 16 bytes; check for success */
void *ckalloc(size_t amount, const char *msg)
{
    void *p;

#if defined MULTI_THREADING && defined MULTI_PROCESSOR
    /* There is a nasty bug in MPI-GM that takes over malloc/free:
     * their routines are not thread-safe and interfere with other
     * MPI calls.  Therefore, protect malloc and free with the MPI
     * mutex. */

    pthread_mutex_lock(&comm_mutex);
#endif

#if defined __INTEL_COMPILER && (defined SSE || defined SSE2)
    p = _mm_malloc(amount, sizeof(block_type));
#else
    p = malloc(amount);
#endif

    if (p == 0) {
	fprintf(stderr, "Ran out of memory at allocating: %s.\n", msg);
	exit(1);
    }

#if defined MULTI_THREADING && defined MULTI_PROCESSOR
    pthread_mutex_unlock(&comm_mutex);
#endif

    return p;
}

/* ckcalloc - allocate space and init to zero; check for success */
void *ckcalloc(size_t amount, size_t len, const char *msg)
{
    void   *p;
    size_t size = amount * len;

    p = ckalloc(size, msg);
    memset(p, 0, size);

    return p;
}


void ckfree(void *p)
{
#if defined MULTI_THREADING && defined MULTI_PROCESSOR
    pthread_mutex_lock(&comm_mutex);
#endif

#if defined __INTEL_COMPILER && (defined SSE || defined SSE2)
   _mm_free(p);
#else
    free(p);
#endif

#if defined MULTI_THREADING && defined MULTI_PROCESSOR
    pthread_mutex_unlock(&comm_mutex);
#endif
}
